import math
from collections import defaultdict
from typing import Dict, Set, Iterator, Tuple, FrozenSet

import typing

from randovania.game_description.game_description import calculate_interesting_resources
from randovania.game_description.node import ResourceNode, Node
from randovania.game_description.requirements import RequirementList, RequirementSet, SatisfiableRequirements, \
    RequirementAnd, Requirement
from randovania.resolver import debug
from randovania.resolver.logic import Logic
from randovania.resolver.state import State


class ResolverReach:
    _nodes: Tuple[Node, ...]
    _energy_at_node: Dict[Node, int]
    path_to_node: Dict[Node, Tuple[Node, ...]]
    _satisfiable_requirements: SatisfiableRequirements
    _safe_nodes: FrozenSet[Node]
    _logic: Logic

    @property
    def nodes(self) -> Iterator[Node]:
        return iter(self._nodes)

    @property
    def satisfiable_requirements(self) -> SatisfiableRequirements:
        return self._satisfiable_requirements

    @property
    def satisfiable_as_requirement_set(self) -> RequirementSet:
        return RequirementSet(self._satisfiable_requirements)

    def __init__(self,
                 nodes: Dict[Node, int],
                 path_to_node: Dict[Node, Tuple[Node, ...]],
                 requirements: SatisfiableRequirements,
                 logic: Logic):
        self._nodes = tuple(nodes.keys())
        self._energy_at_node = nodes
        self._logic = logic
        self.path_to_node = path_to_node
        self._satisfiable_requirements = requirements

    @classmethod
    def calculate_reach(cls,
                        logic: Logic,
                        initial_state: State) -> "ResolverReach":

        checked_nodes: Dict[Node, int] = {}

        # Keys: nodes to check
        # Value: how much energy was available when visiting that node
        nodes_to_check: Dict[Node, int] = {
            initial_state.node: initial_state.energy
        }

        reach_nodes: Dict[Node, int] = {}
        requirements_by_node: Dict[Node, Set[RequirementList]] = defaultdict(set)

        path_to_node: Dict[Node, Tuple[Node, ...]] = {}
        path_to_node[initial_state.node] = tuple()

        while nodes_to_check:
            node = next(iter(nodes_to_check))
            energy = nodes_to_check.pop(node)

            if node.heal:
                energy = initial_state.maximum_energy

            checked_nodes[node] = energy
            if node != initial_state.node:
                reach_nodes[node] = energy

            requirement_to_leave = node.requirement_to_leave(initial_state.patches, initial_state.resources)

            for target_node, requirement in logic.game.world_list.potential_nodes_from(node, initial_state.patches):
                if target_node is None:
                    continue

                if checked_nodes.get(target_node, math.inf) <= energy or nodes_to_check.get(target_node,
                                                                                            math.inf) <= energy:
                    continue

                if requirement_to_leave != Requirement.trivial():
                    requirement = RequirementAnd([requirement, requirement_to_leave])

                # Check if the normal requirements to reach that node is satisfied
                satisfied = requirement.satisfied(initial_state.resources, energy)
                if satisfied:
                    # If it is, check if we additional requirements figured out by backtracking is satisfied
                    satisfied = logic.get_additional_requirements(node).satisfied(initial_state.resources,
                                                                                  energy)

                if satisfied:
                    nodes_to_check[target_node] = energy - requirement.damage(initial_state.resources)
                    path_to_node[target_node] = path_to_node[node] + (node,)

                elif target_node:
                    # If we can't go to this node, store the reason in order to build the satisfiable requirements.
                    # Note we ignore the 'additional requirements' here because it'll be added on the end.
                    requirements_by_node[target_node].update(requirement.as_set.alternatives)

        # Discard satisfiable requirements of nodes reachable by other means
        for node in set(reach_nodes.keys()).intersection(requirements_by_node.keys()):
            requirements_by_node.pop(node)

        if requirements_by_node:
            satisfiable_requirements = frozenset.union(
                *[RequirementSet(requirements).union(logic.get_additional_requirements(node)).alternatives
                  for node, requirements in requirements_by_node.items()])
        else:
            satisfiable_requirements = frozenset()

        return ResolverReach(reach_nodes, path_to_node,
                             satisfiable_requirements,
                             logic)

    def possible_actions(self,
                         state: State) -> Iterator[Tuple[ResourceNode, int]]:

        for node in self.collectable_resource_nodes(state):
            additional_requirements = self._logic.get_additional_requirements(node)
            if additional_requirements.satisfied(state.resources, self._energy_at_node[node]):
                yield node, self._energy_at_node[node]
            else:
                debug.log_skip_action_missing_requirement(node, self._logic.game,
                                                          self._logic.get_additional_requirements(node))

    def satisfiable_actions(self,
                            state: State,
                            victory_condition: Requirement,
                            ) -> Iterator[Tuple[ResourceNode, int]]:

        # print(" > interesting_resources from {} satisfiable_requirements".format(len(satisfiable_requirements)))
        interesting_resources = calculate_interesting_resources(
            self._satisfiable_requirements.union(victory_condition.as_set.alternatives),
            state.resources,
            state.energy,
            state.resource_database)

        # print(" > satisfiable actions, with {} interesting resources".format(len(interesting_resources)))
        for action, energy in self.possible_actions(state):
            for resource, amount in action.resource_gain_on_collect(state.patches, state.resources,
                                                                    self._logic.game.world_list.all_nodes):
                if resource in interesting_resources:
                    yield action, energy
                    break

    def collectable_resource_nodes(self,
                                   state: State) -> Iterator[ResourceNode]:
        for node in self.nodes:
            if not node.is_resource_node:
                continue
            node = typing.cast(ResourceNode, node)
            if node.can_collect(state.patches, state.resources, self._logic.game.world_list.all_nodes):
                yield node



import re
from distutils.version import StrictVersion
from typing import NamedTuple, List, Tuple, Optional

from randovania import VERSION


class VersionDescription(NamedTuple):
    tag_name: str
    change_log: str
    html_url: str

    @property
    def as_strict_version(self) -> StrictVersion:
        return StrictVersion(self.tag_name[1:])


def strict_version_for_version_string(version_name: str) -> StrictVersion:
    try:
        return StrictVersion(version_name)
    except ValueError:
        if ".dev" not in version_name:
            raise
        return StrictVersion(version_name.split(".dev")[0])


def strict_current_version() -> StrictVersion:
    return strict_version_for_version_string(VERSION)


def get_version_for_release(release: dict) -> VersionDescription:
    return VersionDescription(release["tag_name"], release["body"], release["html_url"])


def versions_to_display_for_releases(current_version: StrictVersion,
                                     last_changelog_version: StrictVersion,
                                     releases: List[dict],
                                     ) -> Tuple[List[str], List[str], Optional[VersionDescription]]:
    all_change_logs = []
    new_change_logs = []
    displayed_new_version = False
    version_to_display = None

    for release in releases:
        version = get_version_for_release(release)
        strict_version = version.as_strict_version

        if strict_version > current_version:
            if not displayed_new_version:
                version_to_display = version
                displayed_new_version = True

        else:
            log = "## {}\n\n{}".format(version.tag_name, version.change_log)
            all_change_logs.append(log)

            if strict_version > last_changelog_version:
                if "*Major*" in log:
                    first_non_major = re.search(r"\n\s*-\s+[^*]+\n", log)
                    if first_non_major is not None:
                        log = log[:first_non_major.start()] + "\n\n---\nFor more details, check the Change Log tab."
                new_change_logs.append(log)

    return all_change_logs, new_change_logs, version_to_display



from pathlib import Path
from typing import Optional, Tuple


def discover_game(game_files_path: Path) -> Optional[Tuple[str, str]]:
    if not game_files_path.is_dir():
        return None

    boot_bin = game_files_path / "sys" / "boot.bin"
    try:
        header_bytes = boot_bin.read_bytes()

        game_id = header_bytes[0:6].decode("UTF-8")
        game_title = header_bytes[0x20:(0x20 + 40)].split(b"\x00")[0].decode("UTF-8")
        return game_id, game_title

    except FileNotFoundError:
        return None


def validate_game_files_path(game_files_path: Path):
    if not game_files_path.is_dir():
        raise ValueError("Not a directory")

    required_files = ["default.dol", "FrontEnd.pak", "Metroid1.pak", "Metroid2.pak"]
    missing_files = [(game_files_path / required_file).is_file()
                     for required_file in required_files]

    if not all(missing_files):
        raise ValueError("Is not a valid game folder. Missing files: {}".format([
            filename for filename, exists in zip(required_files, missing_files)
            if not exists
        ]))



import functools
from typing import Optional

import flask
import flask_socketio
import peewee
import socketio
from cryptography.fernet import Fernet
from flask_discord import DiscordOAuth2Session
from prometheus_flask_exporter import PrometheusMetrics

from randovania.network_common.error import NotLoggedIn, BaseNetworkError, ServerError, InvalidSession
from randovania.server.database import User, GameSessionMembership
from randovania.server.lib import logger


class ServerApp:
    sio: flask_socketio.SocketIO
    discord: DiscordOAuth2Session
    metrics: PrometheusMetrics
    fernet_encrypt: Fernet
    guest_encrypt: Optional[Fernet] = None

    def __init__(self, app: flask.Flask):
        self.app = app
        self.sio = flask_socketio.SocketIO(app)
        self.discord = DiscordOAuth2Session(app)
        self.metrics = PrometheusMetrics(app)
        self.fernet_encrypt = Fernet(app.config["FERNET_KEY"])
        if app.config["GUEST_KEY"] is not None:
            self.guest_encrypt = Fernet(app.config["GUEST_KEY"])

    def get_server(self) -> socketio.Server:
        return self.sio.server

    def get_session(self, namespace=None):
        return self.get_server().get_session(flask.request.sid, namespace=namespace)

    def session(self, namespace=None):
        return self.get_server().session(flask.request.sid, namespace=namespace)

    def get_current_user(self) -> User:
        try:
            return User.get_by_id(self.get_session()["user-id"])
        except KeyError:
            raise NotLoggedIn()
        except peewee.DoesNotExist:
            raise InvalidSession()

    def join_game_session(self, membership: GameSessionMembership):
        flask_socketio.join_room(f"game-session-{membership.session.id}")
        flask_socketio.join_room(f"game-session-{membership.session.id}-{membership.user.id}")
        with self.session() as sio_session:
            sio_session["current_game_session"] = membership.session.id

    def leave_game_session(self):
        with self.session() as sio_session:
            if "current_game_session" not in sio_session:
                return
            current_game_session = sio_session.pop("current_game_session")

        user = self.get_current_user()
        flask_socketio.leave_room(f"game-session-{current_game_session}")
        flask_socketio.leave_room(f"game-session-{current_game_session}-{user.id}")

    def on(self, message: str, handler, namespace=None):
        @functools.wraps(handler)
        def _handler(*args):
            try:
                return {
                    "result": handler(self, *args),
                }
            except BaseNetworkError as error:
                return error.as_json

            except Exception:
                logger().exception(f"Unhandled exception while processing request for message {message}. Args: {args}")
                return ServerError().as_json

        metric_wrapper = self.metrics.summary(f"socket_{message}", f"Socket.io messages of type {message}")

        return self.sio.on(message, namespace)(metric_wrapper(_handler))



import base64
import binascii
import json
import operator
from dataclasses import dataclass
from typing import Iterator, Tuple, Dict, Iterable

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder, BitPackValue, single_byte_hash
from randovania.interface_common.preset_manager import PresetManager
from randovania.layout.preset import Preset

_PERMALINK_MAX_VERSION = 16
_PERMALINK_MAX_SEED = 2 ** 31
_PERMALINK_PLAYER_COUNT_LIMITS = (2, 256)


def _dictionary_byte_hash(data: dict) -> int:
    return single_byte_hash(json.dumps(data, separators=(',', ':')).encode("UTF-8"))


def rotate_bytes(data: Iterable[int], rotation: int, per_byte_adjustment: int,
                 inverse: bool = False) -> Iterator[int]:
    """
    Rotates the elements in data, in a reversible operation.
    :param data: The byte values to rotate. Should be ints in the [0, 255] range.
    :param rotation: By how much each item should be rotated, mod 256.
    :param per_byte_adjustment: Increment the rotation by this after each byte, mod 256.
    :param inverse: If True, it performs the inverse operation.
    :return:
    """
    if inverse:
        op = operator.sub
    else:
        op = operator.add
    for b in data:
        yield op(b, rotation) % 256
        rotation = (rotation + per_byte_adjustment) % 256


@dataclass(frozen=True)
class Permalink(BitPackValue):
    seed_number: int
    spoiler: bool
    presets: Dict[int, Preset]

    def __post_init__(self):
        if self.seed_number is None:
            raise ValueError("Missing seed number")
        if not (0 <= self.seed_number < _PERMALINK_MAX_SEED):
            raise ValueError("Invalid seed number: {}".format(self.seed_number))
        object.__setattr__(self, "__cached_as_bytes", None)

    @classmethod
    def current_version(cls) -> int:
        # When this reaches _PERMALINK_MAX_VERSION, we need to change how we decode to avoid breaking version detection
        # for previous Randovania versions
        return 12

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        yield self.current_version(), _PERMALINK_MAX_VERSION
        yield self.seed_number, _PERMALINK_MAX_SEED
        yield from bitpacking.encode_bool(self.spoiler)
        yield from bitpacking.encode_int_with_limits(self.player_count, _PERMALINK_PLAYER_COUNT_LIMITS)

        manager = PresetManager(None)

        previous_unique_presets = []
        for preset in self.presets.values():
            already_encoded_preset = preset in previous_unique_presets
            yield from bitpacking.encode_bool(already_encoded_preset)
            if already_encoded_preset:
                yield from bitpacking.pack_array_element(preset, previous_unique_presets)
            else:
                previous_unique_presets.append(preset)
                yield from preset.bit_pack_encode({"manager": manager})

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata) -> "Permalink":
        version, seed_number = decoder.decode(_PERMALINK_MAX_VERSION, _PERMALINK_MAX_SEED)
        cls._raise_if_different_version(version)

        spoiler = bitpacking.decode_bool(decoder)
        player_count = bitpacking.decode_int_with_limits(decoder, _PERMALINK_PLAYER_COUNT_LIMITS)
        manager = PresetManager(None)

        previous_unique_presets = []
        presets = {}

        for index in range(player_count):
            in_previous_presets = bitpacking.decode_bool(decoder)
            if in_previous_presets:
                preset = decoder.decode_element(previous_unique_presets)
            else:
                preset = Preset.bit_pack_unpack(decoder, {"manager": manager})
                previous_unique_presets.append(preset)
            presets[index] = preset

        return Permalink(seed_number, spoiler, presets)

    @classmethod
    def _raise_if_different_version(cls, version: int):
        if version != cls.current_version():
            raise ValueError("Given permalink has version {}, but this Randovania "
                             "support only permalink of version {}.".format(version, cls.current_version()))

    @classmethod
    def validate_version(cls, b: bytes):
        version = BitPackDecoder(b).peek(_PERMALINK_MAX_VERSION)[0]
        cls._raise_if_different_version(version)

    @property
    def as_bytes(self) -> bytes:
        cached_result = object.__getattribute__(self, "__cached_as_bytes")
        if cached_result is not None:
            return cached_result

        encoded = bitpacking.pack_value(self)
        # Add extra bytes so the base64 encoding never uses == at the end
        encoded += b"\x00" * (3 - (len(encoded) + 1) % 3)

        # Rotate bytes, so the slightest change causes a cascading effect.
        # But skip the first byte so the version check can be done early in decoding.
        byte_hash = single_byte_hash(encoded)
        new_bytes = [encoded[0]]
        new_bytes.extend(rotate_bytes(encoded[1:], byte_hash, byte_hash, inverse=False))

        # Append the hash, so the rotation can be reversed and the checksum verified
        new_bytes.append(byte_hash)

        result = bytes(new_bytes)
        object.__setattr__(self, "__cached_as_bytes", result)
        return result

    @classmethod
    def from_bytes(cls, b: bytes) -> "Permalink":
        Permalink.validate_version(b)

        byte_hash = b[-1]
        new_bytes = [b[0]]
        new_bytes.extend(rotate_bytes(b[1:-1], byte_hash, byte_hash, inverse=True))

        decoded = bytes(new_bytes)
        if single_byte_hash(decoded) != byte_hash:
            raise ValueError("Incorrect checksum")

        decoder = BitPackDecoder(decoded)
        return Permalink.bit_pack_unpack(decoder, {})

    @property
    def as_base64_str(self) -> str:
        try:
            return base64.b64encode(self.as_bytes, altchars=b'-_').decode("utf-8")
        except ValueError as e:
            return "Unable to create Permalink: {}".format(e)

    @classmethod
    def from_str(cls, param: str) -> "Permalink":
        try:
            b = base64.b64decode(param.encode("utf-8"), altchars=b'-_', validate=True)
            if len(b) < 2:
                raise ValueError("Data too small")

            return cls.from_bytes(b)

        except binascii.Error as e:
            raise ValueError("Unable to base64 decode '{permalink}': {error}".format(
                permalink=param,
                error=e,
            ))

    @property
    def player_count(self) -> int:
        return len(self.presets)

    def get_preset(self, index: int) -> Preset:
        return self.presets[index]



import asyncio
import datetime
import os
import subprocess
import urllib.parse

import aiohttp

from randovania import VERSION


async def post_to_discord():
    git_result = subprocess.run(["git", "show"], check=True, stdout=subprocess.PIPE, text=True).stdout.split("\n")
    commit_hash = git_result[0].split()[1]

    message = ""
    for line in git_result[4:]:
        if line.startswith("diff --git"):
            break
        else:
            message += line
            message += "\n"

    download_base = "https://bintray.com/randovania/randovania/download_file?file_path="
    windows_download_url = download_base + urllib.parse.quote(f"randovania-{VERSION}-windows.7z")

    webhook_data = {
        "embeds": [{
            "color": 0x2ecc71,
            "title": f"Randovania {VERSION}",
            "url": f"https://github.com/randovania/randovania/commit/{commit_hash}",
            "description": message.strip(),
            "fields": [
                {
                    "name": "Windows",
                    "value": f"[Download]({windows_download_url})",
                    "inline": True
                },
            ],
            "timestamp": datetime.datetime.now().isoformat()
        }]
    }
    webhook_url = os.environ["DISCORD_WEBHOOK"]
    async with aiohttp.ClientSession() as session:
        async with session.post(webhook_url, json=webhook_data, raise_for_status=True) as response:
            print(await response.text())


if __name__ == '__main__':
    asyncio.run(post_to_discord())



from peewee import SqliteDatabase

from randovania.server import database


def test_init(tmpdir):
    test_db = SqliteDatabase(':memory:')
    with test_db.bind_ctx(database.all_classes):
        test_db.connect(reuse_if_open=True)
        test_db.create_tables(database.all_classes)



import asyncio
import asyncio.futures
import concurrent.futures
import threading
import traceback
from typing import Optional

from PySide2.QtCore import Signal

from randovania.games.prime import claris_randomizer


class BackgroundTaskMixin:
    progress_update_signal = Signal(str, int)
    background_tasks_button_lock_signal = Signal(bool)
    abort_background_task_requested: bool = False
    _background_thread: Optional[threading.Thread] = None

    def _start_thread_for(self, target):
        claris_randomizer.IO_LOOP = asyncio.get_event_loop()
        self._background_thread = threading.Thread(target=target)
        self._background_thread.start()

    def run_in_background_thread(self, target, starting_message: str):

        last_progress = 0

        def progress_update(message: str, progress: Optional[float]):
            nonlocal last_progress
            if progress is None:
                progress = last_progress
            else:
                last_progress = progress

            if self.abort_background_task_requested:
                self.progress_update_signal.emit("{} - Aborted".format(message), int(progress * 100))
                raise AbortBackgroundTask()
            else:
                self.progress_update_signal.emit(message, int(progress * 100))

        def thread(**_kwargs):
            try:
                target(progress_update=progress_update, **_kwargs)
            except AbortBackgroundTask:
                pass
            except Exception as e:
                traceback.print_exc()
                progress_update("Error: {}".format(e), None)
            finally:
                self._background_thread = None
                self.background_tasks_button_lock_signal.emit(True)

        if self._background_thread:
            raise RuntimeError("Trying to start a new background thread while one exists already.")

        self.abort_background_task_requested = False
        progress_update(starting_message, 0)

        self._start_thread_for(thread)
        self.background_tasks_button_lock_signal.emit(False)

    async def run_in_background_async(self, target, starting_message: str):
        fut = concurrent.futures.Future()

        def work(**_kwargs):
            try:
                fut.set_result(target(**_kwargs))
            except AbortBackgroundTask:
                fut.cancel()
            except Exception as e:
                fut.set_exception(e)

        self.run_in_background_thread(work, starting_message)

        return await asyncio.futures.wrap_future(fut)

    def stop_background_process(self):
        self.abort_background_task_requested = True

    @property
    def has_background_process(self) -> bool:
        return self._background_thread is not None


class AbortBackgroundTask(Exception):
    pass



"""Classes that describes the raw data of a game world."""
import copy
from typing import Iterator, FrozenSet, Dict, Optional

from randovania.game_description.area import Area
from randovania.game_description.area_location import AreaLocation
from randovania.game_description.dock import DockWeaknessDatabase
from randovania.game_description.echoes_game_specific import EchoesGameSpecific
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.node import TeleporterNode
from randovania.game_description.requirements import SatisfiableRequirements, Requirement
from randovania.game_description.resources.damage_resource_info import DamageResourceInfo
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import ResourceInfo, ResourceGainTuple, CurrentResources
from randovania.game_description.world_list import WorldList
from randovania.games.game import RandovaniaGame


def _calculate_dangerous_resources_in_db(db: DockWeaknessDatabase) -> Iterator[ResourceInfo]:
    for list_by_type in db:
        for dock_weakness in list_by_type:
            yield from dock_weakness.requirement.as_set.dangerous_resources


def _calculate_dangerous_resources_in_areas(areas: Iterator[Area]) -> Iterator[ResourceInfo]:
    for area in areas:
        for node in area.nodes:
            for requirement in area.connections[node].values():
                yield from requirement.as_set.dangerous_resources


class GameDescription:
    game: RandovaniaGame
    dock_weakness_database: DockWeaknessDatabase

    resource_database: ResourceDatabase
    game_specific: EchoesGameSpecific
    victory_condition: Requirement
    starting_location: AreaLocation
    initial_states: Dict[str, ResourceGainTuple]
    _dangerous_resources: Optional[FrozenSet[ResourceInfo]] = None
    world_list: WorldList

    def __deepcopy__(self, memodict):
        new_game = GameDescription(
            game=self.game,
            resource_database=self.resource_database,
            game_specific=self.game_specific,
            dock_weakness_database=self.dock_weakness_database,
            world_list=copy.deepcopy(self.world_list, memodict),
            victory_condition=self.victory_condition,
            starting_location=self.starting_location,
            initial_states=copy.copy(self.initial_states),
        )
        new_game._dangerous_resources = self._dangerous_resources
        return new_game

    def __init__(self,
                 game: RandovaniaGame,
                 dock_weakness_database: DockWeaknessDatabase,

                 resource_database: ResourceDatabase,
                 game_specific: EchoesGameSpecific,
                 victory_condition: Requirement,
                 starting_location: AreaLocation,
                 initial_states: Dict[str, ResourceGainTuple],
                 world_list: WorldList,
                 ):
        self.game = game
        self.dock_weakness_database = dock_weakness_database

        self.resource_database = resource_database
        self.game_specific = game_specific
        self.victory_condition = victory_condition
        self.starting_location = starting_location
        self.initial_states = initial_states
        self.world_list = world_list

    def patch_requirements(self, resources, damage_multiplier: float):
        self.world_list.patch_requirements(resources, damage_multiplier)
        self._dangerous_resources = None

    def create_game_patches(self) -> GamePatches:
        elevator_connection = {
            node.teleporter_instance_id: node.default_connection

            for node in self.world_list.all_nodes
            if isinstance(node, TeleporterNode) and node.editable
        }

        return GamePatches(None, {}, elevator_connection, {}, {}, {}, {}, self.starting_location, {},
                           game_specific=self.game_specific)

    @property
    def dangerous_resources(self) -> FrozenSet[ResourceInfo]:
        if self._dangerous_resources is None:
            self._dangerous_resources = frozenset(
                _calculate_dangerous_resources_in_areas(self.world_list.all_areas)) | frozenset(
                _calculate_dangerous_resources_in_db(self.dock_weakness_database))
        return self._dangerous_resources


def _resources_for_damage(resource: DamageResourceInfo, database: ResourceDatabase) -> Iterator[ResourceInfo]:
    yield database.energy_tank
    yield from (reduction.inventory_item for reduction in resource.reductions)


def calculate_interesting_resources(satisfiable_requirements: SatisfiableRequirements,
                                    resources: CurrentResources,
                                    energy: int,
                                    database: ResourceDatabase) -> FrozenSet[ResourceInfo]:
    """A resource is considered interesting if it isn't satisfied and it belongs to any satisfiable RequirementList """

    def helper():
        # For each possible requirement list
        for requirement_list in satisfiable_requirements:
            # If it's not satisfied, there's at least one IndividualRequirement in it that can be collected
            if not requirement_list.satisfied(resources, energy):

                for individual in requirement_list.values():
                    # Ignore those with the `negate` flag. We can't "uncollect" a resource to satisfy these.
                    # Finally, if it's not satisfied then we're interested in collecting it
                    if not individual.negate and not individual.satisfied(resources, energy):
                        if isinstance(individual.resource, DamageResourceInfo):
                            yield from _resources_for_damage(individual.resource, database)
                        else:
                            yield individual.resource

    return frozenset(helper())



from PySide2.QtWidgets import QDialog, QApplication

from randovania.gui.generated.permalink_dialog_ui import Ui_PermalinkDialog
from randovania.gui.lib import common_qt_lib
from randovania.layout.permalink import Permalink


class PermalinkDialog(QDialog, Ui_PermalinkDialog):

    def __init__(self):
        super().__init__()
        self.setupUi(self)
        common_qt_lib.set_default_window_icon(self)

        self.accept_button.setEnabled(False)
        self.accept_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)
        self.paste_button.clicked.connect(self._on_paste_button)

        self.permalink_edit.textChanged.connect(self._on_permalink_changed)
        self.permalink_edit.setFocus()

        self._on_paste_button()

    # Permalink
    def get_permalink_from_field(self) -> Permalink:
        return Permalink.from_str(self.permalink_edit.text())

    def _on_permalink_changed(self, value: str):
        common_qt_lib.set_error_border_stylesheet(self.permalink_edit, False)
        common_qt_lib.set_edit_if_different(self.permalink_edit, self.permalink_edit.text().strip())
        self.accept_button.setEnabled(False)
        self.import_error_label.setText("")

        try:
            # Ignoring return value: we only want to know if it's valid
            new_permalink = self.get_permalink_from_field()
            if new_permalink.as_base64_str != self.permalink_edit.text():
                raise ValueError("Imported permalink is different from text field.")
            self.accept_button.setEnabled(True)
        except ValueError as e:
            self.import_error_label.setText(f"Invalid permalink: {e}")
            common_qt_lib.set_error_border_stylesheet(self.permalink_edit, True)

    def _on_paste_button(self):
        self.permalink_edit.setText(QApplication.clipboard().text())



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'game_session_browser_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_GameSessionBrowserDialog(object):
    def setupUi(self, GameSessionBrowserDialog):
        if not GameSessionBrowserDialog.objectName():
            GameSessionBrowserDialog.setObjectName(u"GameSessionBrowserDialog")
        GameSessionBrowserDialog.resize(694, 455)
        self.verticalLayout = QVBoxLayout(GameSessionBrowserDialog)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.menu_bar = QMenuBar(GameSessionBrowserDialog)
        self.menu_bar.setObjectName(u"menu_bar")

        self.verticalLayout.addWidget(self.menu_bar)

        self.table_widget = QTableWidget(GameSessionBrowserDialog)
        if (self.table_widget.columnCount() < 6):
            self.table_widget.setColumnCount(6)
        __qtablewidgetitem = QTableWidgetItem()
        self.table_widget.setHorizontalHeaderItem(0, __qtablewidgetitem)
        __qtablewidgetitem1 = QTableWidgetItem()
        self.table_widget.setHorizontalHeaderItem(1, __qtablewidgetitem1)
        __qtablewidgetitem2 = QTableWidgetItem()
        self.table_widget.setHorizontalHeaderItem(2, __qtablewidgetitem2)
        __qtablewidgetitem3 = QTableWidgetItem()
        self.table_widget.setHorizontalHeaderItem(3, __qtablewidgetitem3)
        __qtablewidgetitem4 = QTableWidgetItem()
        self.table_widget.setHorizontalHeaderItem(4, __qtablewidgetitem4)
        __qtablewidgetitem5 = QTableWidgetItem()
        self.table_widget.setHorizontalHeaderItem(5, __qtablewidgetitem5)
        self.table_widget.setObjectName(u"table_widget")
        self.table_widget.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.table_widget.setTabKeyNavigation(False)
        self.table_widget.setSelectionMode(QAbstractItemView.SingleSelection)
        self.table_widget.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.table_widget.setSortingEnabled(True)
        self.table_widget.verticalHeader().setVisible(False)

        self.verticalLayout.addWidget(self.table_widget)

        self.filter_group = QGroupBox(GameSessionBrowserDialog)
        self.filter_group.setObjectName(u"filter_group")
        self.gridLayout = QGridLayout(self.filter_group)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.has_password_yes_check = QCheckBox(self.filter_group)
        self.has_password_yes_check.setObjectName(u"has_password_yes_check")
        self.has_password_yes_check.setChecked(True)

        self.gridLayout.addWidget(self.has_password_yes_check, 2, 1, 1, 1)

        self.state_inprogress_check = QCheckBox(self.filter_group)
        self.state_inprogress_check.setObjectName(u"state_inprogress_check")
        self.state_inprogress_check.setChecked(True)

        self.gridLayout.addWidget(self.state_inprogress_check, 3, 2, 1, 1)

        self.has_password_no_check = QCheckBox(self.filter_group)
        self.has_password_no_check.setObjectName(u"has_password_no_check")
        self.has_password_no_check.setChecked(True)

        self.gridLayout.addWidget(self.has_password_no_check, 2, 2, 1, 1)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer, 2, 4, 1, 1)

        self.filter_name_edit = QLineEdit(self.filter_group)
        self.filter_name_edit.setObjectName(u"filter_name_edit")

        self.gridLayout.addWidget(self.filter_name_edit, 0, 1, 1, 4)

        self.state_filter_label = QLabel(self.filter_group)
        self.state_filter_label.setObjectName(u"state_filter_label")

        self.gridLayout.addWidget(self.state_filter_label, 3, 0, 1, 1)

        self.state_finished_check = QCheckBox(self.filter_group)
        self.state_finished_check.setObjectName(u"state_finished_check")

        self.gridLayout.addWidget(self.state_finished_check, 3, 3, 1, 1)

        self.state_setup_check = QCheckBox(self.filter_group)
        self.state_setup_check.setObjectName(u"state_setup_check")
        self.state_setup_check.setChecked(True)

        self.gridLayout.addWidget(self.state_setup_check, 3, 1, 1, 1)

        self.filter_name_label = QLabel(self.filter_group)
        self.filter_name_label.setObjectName(u"filter_name_label")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.filter_name_label.sizePolicy().hasHeightForWidth())
        self.filter_name_label.setSizePolicy(sizePolicy)

        self.gridLayout.addWidget(self.filter_name_label, 0, 0, 1, 1)

        self.has_password_label = QLabel(self.filter_group)
        self.has_password_label.setObjectName(u"has_password_label")

        self.gridLayout.addWidget(self.has_password_label, 2, 0, 1, 1)

        self.filter_age_label = QLabel(self.filter_group)
        self.filter_age_label.setObjectName(u"filter_age_label")

        self.gridLayout.addWidget(self.filter_age_label, 4, 0, 1, 1)

        self.filter_age_check = QCheckBox(self.filter_group)
        self.filter_age_check.setObjectName(u"filter_age_check")
        self.filter_age_check.setChecked(True)

        self.gridLayout.addWidget(self.filter_age_check, 4, 1, 1, 1)

        self.filter_age_spin = QSpinBox(self.filter_group)
        self.filter_age_spin.setObjectName(u"filter_age_spin")
        self.filter_age_spin.setMaximum(365)
        self.filter_age_spin.setValue(14)

        self.gridLayout.addWidget(self.filter_age_spin, 4, 2, 1, 2)


        self.verticalLayout.addWidget(self.filter_group)

        self.label_layout = QHBoxLayout()
        self.label_layout.setSpacing(6)
        self.label_layout.setObjectName(u"label_layout")
        self.status_label = QLabel(GameSessionBrowserDialog)
        self.status_label.setObjectName(u"status_label")
        self.status_label.setWordWrap(True)

        self.label_layout.addWidget(self.status_label)

        self.server_connection_label = QLabel(GameSessionBrowserDialog)
        self.server_connection_label.setObjectName(u"server_connection_label")
        self.server_connection_label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.label_layout.addWidget(self.server_connection_label)


        self.verticalLayout.addLayout(self.label_layout)

        self.button_box = QDialogButtonBox(GameSessionBrowserDialog)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setMinimumSize(QSize(500, 0))
        self.button_box.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.button_box)


        self.retranslateUi(GameSessionBrowserDialog)

        QMetaObject.connectSlotsByName(GameSessionBrowserDialog)
    # setupUi

    def retranslateUi(self, GameSessionBrowserDialog):
        GameSessionBrowserDialog.setWindowTitle(QCoreApplication.translate("GameSessionBrowserDialog", u"Session Browser", None))
        ___qtablewidgetitem = self.table_widget.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"Name", None));
        ___qtablewidgetitem1 = self.table_widget.horizontalHeaderItem(1)
        ___qtablewidgetitem1.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"State", None));
        ___qtablewidgetitem2 = self.table_widget.horizontalHeaderItem(2)
        ___qtablewidgetitem2.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"Players", None));
        ___qtablewidgetitem3 = self.table_widget.horizontalHeaderItem(3)
        ___qtablewidgetitem3.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"Password?", None));
        ___qtablewidgetitem4 = self.table_widget.horizontalHeaderItem(4)
        ___qtablewidgetitem4.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"Creator", None));
        ___qtablewidgetitem5 = self.table_widget.horizontalHeaderItem(5)
        ___qtablewidgetitem5.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"Creation Date", None));
        self.filter_group.setTitle(QCoreApplication.translate("GameSessionBrowserDialog", u"Filters", None))
        self.has_password_yes_check.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"Yes", None))
        self.state_inprogress_check.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"In-Progress", None))
        self.has_password_no_check.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"No", None))
        self.state_filter_label.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"State:", None))
        self.state_finished_check.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"Finished", None))
        self.state_setup_check.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"Setup", None))
        self.filter_name_label.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"Name", None))
        self.has_password_label.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"Has Password?", None))
        self.filter_age_label.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"Age:", None))
        self.filter_age_check.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"Limit to", None))
        self.filter_age_spin.setSuffix(QCoreApplication.translate("GameSessionBrowserDialog", u" days", None))
        self.status_label.setText("")
        self.server_connection_label.setText(QCoreApplication.translate("GameSessionBrowserDialog", u"Server: Disconnected", None))
    # retranslateUi




import pypresence
from PySide2 import QtWidgets
from PySide2.QtWidgets import QDialog
from asyncqt import asyncSlot

from randovania.gui.generated.login_prompt_dialog_ui import Ui_LoginPromptDialog
from randovania.gui.lib import common_qt_lib, async_dialog
from randovania.gui.lib.qt_network_client import QtNetworkClient, handle_network_errors
from randovania.network_client.game_session import User
from randovania.network_client.network_client import ConnectionState


class LoginPromptDialog(QDialog, Ui_LoginPromptDialog):

    def __init__(self, network_client: QtNetworkClient):
        super().__init__()
        self.setupUi(self)
        common_qt_lib.set_default_window_icon(self)
        self.network_client = network_client

        login_methods = network_client.available_login_methods
        self.guest_button.setVisible("guest" in login_methods)
        self.discord_button.setEnabled("discord" in login_methods)
        self.discord_button.setToolTip(
            ""
            if self.discord_button.isEnabled() else
            "This Randovania build is not configured to login with Discord."
        )
        self.privacy_policy_label.setText(self.privacy_policy_label.text().replace("color:#0000ff;", ""))

        self.button_box.button(QtWidgets.QDialogButtonBox.Reset).setText("Logout")

        # Signals
        self.network_client.ConnectionStateUpdated.connect(self.on_server_connection_state_updated)
        self.network_client.UserChanged.connect(self.on_user_changed)

        self.guest_button.clicked.connect(self.on_login_as_guest_button)
        self.discord_button.clicked.connect(self.on_login_with_discord_button)
        self.button_box.button(QtWidgets.QDialogButtonBox.Ok).clicked.connect(self.on_ok_button)
        self.button_box.button(QtWidgets.QDialogButtonBox.Reset).clicked.connect(self.on_logout_button)

        # Initial update
        self.on_user_changed(network_client.current_user)

    def on_user_changed(self, user: User):
        self.on_server_connection_state_updated(self.network_client.connection_state)
        self.button_box.button(QtWidgets.QDialogButtonBox.Reset).setEnabled(user is not None)

    def on_server_connection_state_updated(self, state: ConnectionState):
        message = f"{state.value}"
        if state == ConnectionState.Connected:
            message += f", logged as {self.network_client.current_user.name}"
        elif self.network_client.has_previous_session():
            message += " (with saved session)"

        self.connection_status_label.setText(message)

    @asyncSlot()
    @handle_network_errors
    async def on_login_as_guest_button(self):
        dialog = QtWidgets.QInputDialog(self)
        dialog.setModal(True)
        dialog.setWindowTitle("Enter guest name")
        dialog.setLabelText("Select a name for the guest account:")
        if await async_dialog.execute_dialog(dialog) != dialog.Accepted:
            return

        await self.network_client.login_as_guest(dialog.textValue())

    @asyncSlot()
    @handle_network_errors
    async def on_login_with_discord_button(self):
        try:
            await self.network_client.login_with_discord()
        except pypresence.exceptions.InvalidPipe:
            await async_dialog.warning(self, "Discord login",
                                       "Login failed. Is Discord running?")
        except pypresence.exceptions.ServerError:
            await async_dialog.warning(self, "Discord login",
                                       "Login failed. Did you reject the authorization prompt in Discord?")

    def on_ok_button(self):
        self.accept()

    @asyncSlot()
    @handle_network_errors
    async def on_logout_button(self):
        await self.network_client.logout()



import dataclasses
from typing import Tuple, Iterator
from unittest.mock import MagicMock, call

import pytest

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder


@pytest.mark.parametrize("value", [
    10,
    65,
    0,
    1,
    2,
    134,
])
@pytest.mark.parametrize("limits", [
    (5, 20, 500),
    (3, 50, 150),
])
def test_encode_int_with_limits_round_trip(value: int,
                                           limits: Tuple[int, ...],
                                           ):
    # Run
    data = bitpacking._pack_encode_results(list(bitpacking.encode_int_with_limits(value, limits)))
    decoded = bitpacking.decode_int_with_limits(bitpacking.BitPackDecoder(data), limits)

    # Assert
    assert decoded == value


@pytest.fixture(
    params=[
        (0, (1, 4), [(0, 2)]),
        (1, (1, 4), [(1, 2), (0, 4)]),
        (5, (10,), [(5, 11)]),
        (50, (10,), [(50, 11)]),
        (5, (10, 100), [(5, 11)]),
        (50, (10, 100), [(10, 11), (40, 91)]),
        (5, (10, 100, 500), [(5, 11)]),
        (50, (10, 100, 500), [(10, 11), (40, 91)]),
        (500, (10, 100, 500), [(10, 11), (90, 91), (400, 401)]),
    ],
    name="limits_fixture")
def _limits_fixture(request):
    return request.param[0], request.param[1], request.param[2]


def test_encode_int_with_limits(limits_fixture):
    # Setup
    value, limits, encoded = limits_fixture

    # Run
    result = list(bitpacking.encode_int_with_limits(value, limits))

    # Assert
    assert result == encoded


def test_decode_int_with_limits(limits_fixture):
    # Setup
    value, limits, encoded = limits_fixture
    decoder = MagicMock()
    decoder.decode_single.side_effect = [part for part, _ in encoded]

    # Run
    result = bitpacking.decode_int_with_limits(decoder, limits)

    # Assert
    decoder.decode_single.assert_has_calls([
        call(limit)
        for _, limit in encoded
    ])
    assert result == value


@pytest.mark.parametrize(["value", "limits", "expected"], [
    (0, (1, 4), "u1"),
    (1, (1, 4), "u1u2"),
    (2, (1, 4), "u1u2"),
    (3, (1, 4), "u1u2"),
    (4, (1, 4), "u1u2"),
])
def test_encode_int_with_limits_bitstring(value, limits, expected):
    # Run
    result = bitpacking._format_string_for(list(bitpacking.encode_int_with_limits(value, limits)))

    # Assert
    assert result == expected


@pytest.fixture(
    params=[
        (False, (0, 2)),
        (True, (1, 2)),
    ],
    name="bool_fixture")
def _bool_fixture(request):
    return request.param[0], request.param[1]


def test_encode_bool(bool_fixture):
    # Setup
    value, encoded = bool_fixture

    # Run
    result = list(bitpacking.encode_bool(value))

    # Assert
    assert result == [encoded]


def test_decode_bool(bool_fixture):
    # Setup
    value, encoded = bool_fixture
    decoder = MagicMock()
    decoder.decode_single.return_value = encoded[0]

    # Run
    result = bitpacking.decode_bool(decoder)

    # Assert
    decoder.decode_single.assert_called_once_with(encoded[1])
    assert result == value


@pytest.mark.parametrize(["value", "metadata"], [
    (0.0, {"min": 0.0, "max": 1.0, "precision": 1}),
    (0.0, {"min": -1.0, "max": 1.0, "precision": 1}),
    (-0.5, {"min": -1.0, "max": 1.0, "precision": 1}),
    (1.0, {"min": 0.0, "max": 1.0, "precision": 1}),
    (1.0, {"min": 0.0, "max": 1.0, "precision": 2}),
])
def test_round_trip_float(value: float, metadata: dict):
    result = bitpacking.round_trip(bitpacking.BitPackFloat(value), metadata)
    assert result == value


@pytest.mark.parametrize(["elements", "array"], [
    ([], [10, 20]),
    ([10], [10, 20]),
    ([10, 20], [10, 20]),
    ([10, 20], [10, 20, 30]),
    ([10, 20], [10, 20, 30, 50]),
    (list(range(15)), list(range(100))),
    ([x * 2 for x in range(150)], list(range(300))),
])
def test_sorted_array_elements_round_trip(elements, array):
    generator = bitpacking.pack_sorted_array_elements(elements, array)
    b = bitpacking._pack_encode_results(list(generator))
    decoder = bitpacking.BitPackDecoder(b)

    decoded_elements = bitpacking.decode_sorted_array_elements(decoder, array)

    assert elements == decoded_elements


@pytest.mark.parametrize(["elements", "array", "expected_size"], [
    ([], [], 0),
    ([], range(100), 8),
    ([90], range(100), 18),
    (range(100), range(100), 8),
    (range(100), range(300), 219),
    (list(range(100)) + list(range(200, 300)), range(300), 318),
    (range(200), range(300), 120),
    (range(200, 300), range(300), 120),
    ([x * 2 for x in range(150)], range(300), 458),
    ([x * 3 for x in range(100)], range(300), 310),
])
def test_sorted_array_elements_size(elements, array, expected_size):
    count = 0
    for _, size in bitpacking.pack_sorted_array_elements(list(elements), list(array)):
        count += bitpacking._bits_for_number(size)
    assert count == expected_size


def test_pack_array_element_missing():
    with pytest.raises(ValueError):
        list(bitpacking.pack_array_element(5, [10, 25]))


def test_pack_array_element_single():
    assert len(list(bitpacking.pack_array_element("x", ["x"]))) == 0


@pytest.mark.parametrize(["element", "array"], [
    (10, [10, 20]),
    ("x", [10, "x", 20]),
    ("x", ["x"]),
])
def test_array_elements_round_trip(element, array):
    generator = bitpacking.pack_array_element(element, array)
    b = bitpacking._pack_encode_results(list(generator))
    decoder = bitpacking.BitPackDecoder(b)

    decoded_element = decoder.decode_element(array)

    assert element == decoded_element


class BitPackValueUsingReference(bitpacking.BitPackValue):
    value: int

    def __init__(self, x):
        self.value = x

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        reference: BitPackValueUsingReference = metadata["reference"]
        yield self.value - reference.value, 128

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata):
        reference: BitPackValueUsingReference = metadata["reference"]
        value = decoder.decode_single(128) + reference.value
        return BitPackValueUsingReference(value)

    def __eq__(self, other):
        return self.value == other.value


@dataclasses.dataclass(frozen=True)
class DataclassForTest(bitpacking.BitPackDataClass):
    uses_reference: BitPackValueUsingReference


@pytest.mark.parametrize(["value", "reference"], [
    (10, 5),
    (20, 5),
    (50, 5),
])
def test_round_trip_dataclass_for_test(value, reference):
    data = DataclassForTest(BitPackValueUsingReference(value))
    ref = DataclassForTest(BitPackValueUsingReference(reference))

    result = bitpacking.round_trip(data, {"reference": ref})
    assert result == data



"""
Some constants related to specific Prime 1 items that the logic needs to deal with.
"""

ARTIFACT_MODEL = [29 + i for i in range(12)]
ARTIFACT_ITEMS = [29 + i for i in range(12)]
ARTIFACT_NAMES = [
    "Artifact of Truth",
    "Artifact of Strength",
    "Artifact of Elder",
    "Artifact of Wild",
    "Artifact of Lifegiver",
    "Artifact of Warrior",
    "Artifact of Chozo",
    "Artifact of Nature",
    "Artifact of Sun",
    "Artifact of World",
    "Artifact of Spirit",
    "Artifact of Newborn"
]



import copy
from random import Random
from typing import Tuple, Iterator, Callable

from randovania.game_description.assignment import PickupAssignment
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.node import Node, PickupNode
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.generator.filler.filler_library import filter_unassigned_pickup_nodes
from randovania.generator.generator_reach import advance_reach_with_possible_unsafe_resources, \
    reach_with_all_safe_resources, collect_all_safe_resources_in_reach, filter_reachable, pickup_nodes_that_can_reach
from randovania.resolver import debug
from randovania.resolver.logic import Logic
from randovania.resolver.random_lib import iterate_with_weights
from randovania.resolver.state import State, state_with_pickup, add_pickup_to_state


def random_assumed_filler(logic: Logic,
                          initial_state: State,
                          patches: GamePatches,
                          available_pickups: Tuple[PickupEntry],
                          rng: Random,
                          status_update: Callable[[str], None],
                          ) -> PickupAssignment:
    pickup_assignment = copy.copy(patches.pickup_assignment)
    print("Major items: {}".format([item.name for item in available_pickups]))
    game = logic.game

    base_reach = advance_reach_with_possible_unsafe_resources(reach_with_all_safe_resources(game, initial_state))

    reaches_for_pickup = {}

    previous_reach = base_reach
    for pickup in reversed(available_pickups):
        print("** Preparing reach for {}".format(pickup.name))
        new_reach = copy.deepcopy(previous_reach)
        add_pickup_to_state(new_reach.state, pickup)
        new_reach.state.previous_state = new_reach.state
        new_reach.advance_to(new_reach.state)
        collect_all_safe_resources_in_reach(new_reach)
        previous_reach = advance_reach_with_possible_unsafe_resources(new_reach)
        reaches_for_pickup[pickup] = previous_reach

    for i, pickup in enumerate(available_pickups):
        print("\n\n\nWill place {}, have {} pickups left".format(pickup, len(available_pickups) - i - 1))
        reach = reaches_for_pickup[pickup]
        debug.print_actions_of_reach(reach)
        escape_state = state_with_pickup(reach.state, pickup)

        total_pickup_nodes = list(_filter_pickups(filter_reachable(reach.nodes, reach)))
        pickup_nodes = list(filter_unassigned_pickup_nodes(total_pickup_nodes, pickup_assignment))
        num_nodes = len(pickup_nodes)
        actions_weights = {
            node: len(path)
            for node, path in reach.shortest_path_from(initial_state.node).items()
        }

        try:
            pickup_node = next(pickup_nodes_that_can_reach(iterate_with_weights(pickup_nodes, actions_weights, rng),
                                                           reach_with_all_safe_resources(game, escape_state),
                                                           set(reach.safe_nodes)))
            print("Placed {} at {}. Had {} available of {} nodes.".format(pickup.name,
                                                                          game.world_list.node_name(pickup_node, True),
                                                                          num_nodes,
                                                                          len(total_pickup_nodes)))

        except StopIteration:
            print("\n".join(game.world_list.node_name(node, True) for node in reach.safe_nodes))
            raise Exception("Couldn't place {}. Had {} available of {} nodes.".format(pickup.name,
                                                                                      num_nodes,
                                                                                      len(total_pickup_nodes)
                                                                                      ))

        pickup_assignment[pickup_node.pickup_index] = pickup

    return pickup_assignment


def _filter_pickups(nodes: Iterator[Node]) -> Iterator[PickupNode]:
    return filter(lambda node: isinstance(node, PickupNode), nodes)



import io
from unittest.mock import patch, MagicMock

import pytest

import randovania
from randovania.game_description import pretty_print
from randovania.game_description.requirements import RequirementAnd, ResourceRequirement, RequirementTemplate


@pytest.mark.skipif(randovania.is_frozen(), reason="frozen executable doesn't have JSON files")
def test_commited_human_readable_description(echoes_game_description):
    buffer = io.StringIO()
    pretty_print.write_human_readable_world_list(echoes_game_description, buffer)

    assert randovania.get_data_path().joinpath("json_data", "prime2.txt").read_text("utf-8") == buffer.getvalue()


@patch("randovania.game_description.pretty_print.pretty_print_requirement")
def test_pretty_print_requirement_array_one_item(mock_print_requirement: MagicMock):
    mock_print_requirement.return_value = ["a", "b"]

    req = MagicMock()
    array = RequirementAnd([req])

    # Run
    result = list(pretty_print.pretty_print_requirement_array(array, 3))

    # Assert
    assert result == ["a", "b"]
    mock_print_requirement.assert_called_once_with(req, 3)


@patch("randovania.game_description.pretty_print.pretty_print_requirement")
def test_pretty_print_requirement_array_combinable(mock_print_requirement: MagicMock, echoes_resource_database):
    mock_print_requirement.return_value = ["a", "b"]

    array = RequirementAnd([
        ResourceRequirement(echoes_resource_database.item[0], 1, False),
        RequirementTemplate(echoes_resource_database, "Shoot Sunburst"),
    ])

    # Run
    result = list(pretty_print.pretty_print_requirement_array(array, 3))

    # Assert
    assert result == [(3, "Power Beam and Shoot Sunburst")]
    mock_print_requirement.assert_not_called()


import datetime
import json
import logging
from pathlib import Path
from typing import Optional, List

import aiofiles
import aiohttp

from randovania.interface_common import persistence

_RELEASES_URL = "https://api.github.com/repos/randovania/randovania/releases"


def _last_check_file() -> Path:
    return persistence.user_data_dir() / "last_releases.json"


def _is_recent(last_check) -> bool:
    last_check_date = datetime.datetime.fromisoformat(last_check["last_check"])
    return (datetime.datetime.now() - last_check_date) <= datetime.timedelta(days=1)


async def _read_from_persisted() -> Optional[List[dict]]:
    try:
        async with aiofiles.open(_last_check_file()) as open_file:
            last_check = json.loads(await open_file.read())

        if _is_recent(last_check):
            return last_check["data"]
        else:
            return None

    except json.JSONDecodeError as e:
        logging.warning("Unable to decode persisted releases check: %s", str(e))
        return None

    except FileNotFoundError:
        return None


async def _download_from_github() -> Optional[List[dict]]:
    async with aiohttp.ClientSession() as session:
        async with session.get(_RELEASES_URL) as response:
            try:
                response.raise_for_status()
                return await response.json()
            except aiohttp.ClientResponseError:
                return None


async def _persist(data: List[dict]):
    _last_check_file().parent.mkdir(parents=True, exist_ok=True)

    async with aiofiles.open(_last_check_file(), "w") as open_file:
        await open_file.write(
            json.dumps(
                {
                    "last_check": datetime.datetime.now().isoformat(),
                    "data": data,
                },
                default=str))


async def get_releases() -> Optional[List[dict]]:
    data = await _read_from_persisted()

    if data is None:
        data = await _download_from_github()
        if data is not None:
            await _persist(data)

    return data



import pytest
from mock import MagicMock, AsyncMock, call

import randovania
from randovania.games.game import RandovaniaGame
from randovania.server import bot


@pytest.mark.asyncio
async def test_on_message_from_bot(mocker):
    mock_look_for: AsyncMock = mocker.patch("randovania.server.bot.look_for_permalinks", new_callable=AsyncMock)
    client = bot.Bot(None)

    message = MagicMock()
    message.author = client.user

    # Run
    await client.on_message(message)

    # Assert
    mock_look_for.assert_not_awaited()


@pytest.mark.parametrize(["name_filter", "guild", "expected"], [
    ("unknown", 0, False),
    ("expected_name", 0, False),
    ("unknown", 1234, False),
    ("expected_name", 1234, True),
])
@pytest.mark.asyncio
async def test_on_message_wrong_place(mocker, name_filter: str, guild: int, expected):
    mock_look_for: AsyncMock = mocker.patch("randovania.server.bot.look_for_permalinks", new_callable=AsyncMock)
    client = bot.Bot({"channel_name_filter": name_filter, "guild": guild})

    message = MagicMock()
    message.channel.name = "the_expected_name"
    message.guild.id = 1234

    # Run
    await client.on_message(message)

    # Assert
    if expected:
        mock_look_for.assert_awaited_once_with(message.content, message.channel)
    else:
        mock_look_for.assert_not_awaited()


@pytest.mark.asyncio
async def test_look_for_permalinks(mocker):
    permalink_1 = MagicMock()
    permalink_1.player_count = 2

    preset = MagicMock()
    preset.game = RandovaniaGame.PRIME2
    permalink_2 = MagicMock()
    permalink_2.player_count = 1
    permalink_2.get_preset.return_value = preset

    embed_1 = MagicMock()
    embed_2 = MagicMock()

    mock_embed: MagicMock = mocker.patch("discord.Embed", side_effect=[embed_1, embed_2])

    mock_describe: MagicMock = mocker.patch("randovania.gui.lib.preset_describer.describe",
                                            return_value=[
                                                ("General", ["Foo", "Bar"]),
                                                ("Other", ["X", "Y"]),
                                            ])
    mock_from_str: MagicMock = mocker.patch("randovania.layout.permalink.Permalink.from_str",
                                            side_effect=[permalink_1, permalink_2])

    channel = AsyncMock()
    message = "yu4abbceWfLI- `yua73123yWdLI-` foo???"

    # Run
    await bot.look_for_permalinks(message, channel)

    # Assert
    mock_from_str.assert_has_calls([
        call("yu4abbceWfLI-"),
        call("yua73123yWdLI-"),
    ])
    permalink_2.get_preset.assert_called_once_with(0)
    mock_describe.assert_called_once_with(preset)

    mock_embed.assert_has_calls([
        call(title="yu4abbceWfLI-", description="2 player multiworld permalink"),
        call(title="yua73123yWdLI-", description="1 player multiworld permalink"),
    ])
    embed_1.add_field.assert_not_called()
    assert embed_2.description == "Metroid Prime 2: Echoes permalink for Randovania {}".format(randovania.VERSION)
    embed_2.add_field.assert_has_calls([
        call(name="General", value="Foo\nBar", inline=True),
        call(name="Other", value="X\nY", inline=True),
    ])
    channel.send.assert_has_awaits([
        call(embed=embed_1),
        call(embed=embed_2),
    ])



import dataclasses
from dataclasses import dataclass
from unittest.mock import patch, MagicMock

import pytest

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder, BitPackValue
from randovania.layout.ammo_configuration import AmmoConfiguration
from randovania.layout.available_locations import AvailableLocationsConfiguration
from randovania.layout.beam_configuration import BeamConfiguration
from randovania.layout.hint_configuration import HintConfiguration
from randovania.layout.echoes_configuration import EchoesConfiguration, LayoutSkyTempleKeyMode
from randovania.layout.elevators import LayoutElevators
from randovania.layout.damage_strictness import LayoutDamageStrictness
from randovania.layout.major_items_configuration import MajorItemsConfiguration
from randovania.layout.starting_location import StartingLocation
from randovania.layout.translator_configuration import TranslatorConfiguration
from randovania.layout.trick_level import TrickLevelConfiguration


@dataclass(frozen=True)
class DummyValue(BitPackValue):
    def bit_pack_encode(self, metadata):
        yield from []

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata):
        raise cls()


@pytest.fixture(
    params=[
        {"encoded": b'@X=\xef\x898\x0f\x00',
         "sky_temple": LayoutSkyTempleKeyMode.NINE,
         "elevators": LayoutElevators.VANILLA,
         },
        {"encoded": b'@\x00=\xef\x898\x0f\x00',
         "sky_temple": LayoutSkyTempleKeyMode.ALL_BOSSES,
         "elevators": LayoutElevators.VANILLA,
         },
        {"encoded": b'@\xa0=\xef\x898\x0f\x00',
         "sky_temple": LayoutSkyTempleKeyMode.TWO,
         "elevators": LayoutElevators.TWO_WAY_RANDOMIZED,
         },
        {"encoded": b'@\x88=\xef\x898\x0f\x00',
         "sky_temple": LayoutSkyTempleKeyMode.ALL_GUARDIANS,
         "elevators": LayoutElevators.TWO_WAY_RANDOMIZED,
         },
        {"encoded": b'\x00\x88=\xef\x898\x0f\x00',
         "sky_temple": LayoutSkyTempleKeyMode.ALL_GUARDIANS,
         "elevators": LayoutElevators.TWO_WAY_RANDOMIZED,
         "damage_strictness": LayoutDamageStrictness.STRICT,
         },
    ],
    name="layout_config_with_data")
def _layout_config_with_data(request, default_layout_configuration):
    trick_config = DummyValue()
    starting_location = DummyValue()
    available_locations = DummyValue()
    major_items = DummyValue()
    ammo_config = DummyValue()
    translator_config = DummyValue()
    hints = DummyValue()
    beam_configuration = DummyValue()

    with patch.multiple(TrickLevelConfiguration, bit_pack_unpack=MagicMock(return_value=trick_config)), \
         patch.multiple(StartingLocation, bit_pack_unpack=MagicMock(return_value=starting_location)), \
         patch.multiple(AvailableLocationsConfiguration, bit_pack_unpack=MagicMock(return_value=available_locations)), \
         patch.multiple(MajorItemsConfiguration, bit_pack_unpack=MagicMock(return_value=major_items)), \
         patch.multiple(AmmoConfiguration, bit_pack_unpack=MagicMock(return_value=ammo_config)), \
         patch.multiple(TranslatorConfiguration, bit_pack_unpack=MagicMock(return_value=translator_config)), \
         patch.multiple(HintConfiguration, bit_pack_unpack=MagicMock(return_value=hints)), \
         patch.multiple(BeamConfiguration, bit_pack_unpack=MagicMock(return_value=beam_configuration)):
        yield request.param["encoded"], dataclasses.replace(
            default_layout_configuration,
            trick_level=trick_config,
            damage_strictness=request.param.get("damage_strictness", LayoutDamageStrictness.MEDIUM),
            sky_temple_keys=request.param["sky_temple"],
            elevators=request.param["elevators"],
            starting_location=starting_location,
            available_locations=available_locations,
            major_items_configuration=major_items,
            ammo_configuration=ammo_config,
            translator_configuration=translator_config,
            hints=hints,
            beam_configuration=beam_configuration,
        )


def test_decode(layout_config_with_data):
    # Setup
    data, expected = layout_config_with_data

    # Run
    decoder = BitPackDecoder(data)
    result = EchoesConfiguration.bit_pack_unpack(decoder, {})

    # Assert
    assert result == expected


def test_encode(layout_config_with_data):
    # Setup
    expected, value = layout_config_with_data

    # Run
    result = bitpacking.pack_value(value)

    # Assert
    assert result == expected



import time
from argparse import ArgumentParser
from pathlib import Path

from randovania.cli.echoes_lib import add_debug_argument
from randovania.layout.layout_description import LayoutDescription
from randovania.resolver import debug, resolver


def validate_command_logic(args):
    debug.set_level(args.debug)

    description = LayoutDescription.from_file(args.layout_file)

    if description.permalink.player_count != 1:
        raise ValueError(f"Validator does not support layouts with more than 1 player.")

    configuration = description.permalink.presets[0].configuration
    patches = description.all_patches[0]

    before = time.perf_counter()
    final_state_by_resolve = resolver.resolve(
        configuration=configuration,
        patches=patches
    )
    after = time.perf_counter()
    print("Took {} seconds. Game is {}.".format(
        after - before,
        "possible" if final_state_by_resolve is not None else "impossible")
    )


def add_validate_command(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "validate",
        help="Validate a pickup distribution."
    )

    add_debug_argument(parser)
    parser.add_argument(
        "layout_file",
        type=Path,
        help="The layout seed log file to validate.")
    parser.set_defaults(func=validate_command_logic)



from unittest.mock import MagicMock, ANY

import pytest

from randovania.cli.commands import randomize_command
from randovania.interface_common.cosmetic_patches import CosmeticPatches
from randovania.interface_common.players_configuration import PlayersConfiguration


@pytest.mark.parametrize("with_permalink", [False, True])
def test_randomize_command_logic(mocker, with_permalink):
    mock_from_str = mocker.patch("randovania.layout.permalink.Permalink.from_str")
    mock_from_file = mocker.patch("randovania.layout.layout_description.LayoutDescription.from_file")
    mock_generate = mocker.patch("randovania.generator.generator.generate_description")
    mock_apply = mocker.patch("randovania.games.prime.claris_randomizer.apply_layout")

    args = MagicMock()
    if with_permalink:
        layout_description = mock_generate.return_value
    else:
        layout_description = mock_from_file.return_value
        args.permalink = None

    cosmetic_patches = CosmeticPatches(
        disable_hud_popup=args.disable_hud_popup,
        speed_up_credits=args.speed_up_credits)

    # Run
    randomize_command.randomize_command_logic(args)

    # Assert
    mock_apply.assert_called_once_with(
        description=layout_description,
        cosmetic_patches=cosmetic_patches,
        backup_files_path=args.backup_files,
        progress_update=ANY,
        game_root=args.game_files,
        players_config=PlayersConfiguration(0, {0: "You"})
    )

    if with_permalink:
        mock_from_str.assert_called_once_with(args.permalink)
        mock_generate.assert_called_once_with(permalink=mock_from_str.return_value,
                                              status_update=ANY,
                                              validate_after_generation=True)
        mock_from_file.assert_not_called()
    else:
        mock_from_str.assert_not_called()
        mock_generate.assert_not_called()
        mock_from_file.assert_called_once_with(args.log_file)



import asyncio
import base64
import hashlib
import json
import logging
from enum import Enum
from pathlib import Path
from typing import List, Optional, Tuple, Dict, Any

import aiofiles
import aiohttp.client_exceptions
import engineio
import socketio
import socketio.exceptions

import randovania
from randovania.game_connection.backend_choice import GameBackendChoice
from randovania.game_connection.connection_base import InventoryItem, GameConnectionStatus
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.network_client.game_session import GameSessionListEntry, GameSessionEntry, User
from randovania.network_common.admin_actions import SessionAdminUserAction, SessionAdminGlobalAction
from randovania.network_common.error import decode_error, InvalidSession


class ConnectionState(Enum):
    Disconnected = "Disconnected"
    Connecting = "Connecting"
    Connected = "Connected"
    ConnectedNotLogged = "Connected, not logged in"
    ConnectedRestoringSession = "Connected, restoring session"

    @property
    def is_disconnected(self) -> bool:
        return self in (ConnectionState.Disconnected, ConnectionState.Connecting)


def _hash_address(server_address: str) -> str:
    return base64.urlsafe_b64encode(hashlib.blake2b(server_address.encode("utf-8"),
                                                    digest_size=12).digest()).decode("utf-8")


class UnableToConnect(Exception):
    def __init__(self, reason: str):
        super().__init__(reason)
        self.reason = reason


class NetworkClient:
    sio: socketio.AsyncClient
    _current_game_session: Optional[GameSessionEntry] = None
    _current_user: Optional[User] = None
    _connection_state: ConnectionState
    _call_lock: asyncio.Lock
    _connect_task: Optional[asyncio.Task] = None
    _waiting_for_on_connect: Optional[asyncio.Future] = None
    _restore_session_task: Optional[asyncio.Task] = None
    _connect_error: Optional[str] = None
    _last_self_update: Any = None

    def __init__(self, user_data_dir: Path, configuration: dict):
        self.logger = logging.getLogger(__name__)

        self._connection_state = ConnectionState.Disconnected
        self.sio = socketio.AsyncClient()
        self._call_lock = asyncio.Lock()

        self.configuration = configuration
        encoded_address = _hash_address(self.configuration["server_address"])
        self.server_data_path = user_data_dir / encoded_address
        self.session_data_path = self.server_data_path / "session_persistence.bin"

        self.sio.on('connect', self.on_connect)
        self.sio.on('connect_error', self.on_connect_error)
        self.sio.on('disconnect', self.on_disconnect)
        self.sio.on('user_session_update', self.on_user_session_updated)
        self.sio.on('game_session_update', self.on_game_session_updated)
        self.sio.on('game_has_update', self.on_game_update_notification)

    @property
    def connection_state(self) -> ConnectionState:
        return self._connection_state

    @connection_state.setter
    def connection_state(self, value: ConnectionState):
        self.logger.debug(f"updated connection_state: {value.value}")
        self._connection_state = value

    async def read_persisted_session(self) -> Optional[bytes]:
        try:
            async with aiofiles.open(self.session_data_path, "rb") as open_file:
                return await open_file.read()

        except FileNotFoundError:
            return None

    def has_previous_session(self) -> bool:
        return self.session_data_path.is_file()

    async def connect_if_authenticated(self):
        if self.session_data_path.is_file():
            self.logger.debug("session data exists")
            try:
                return await self.connect_to_server()
            except UnableToConnect:
                pass
        else:
            self.logger.debug("no session data")

    async def _internal_connect_to_server(self):
        if self.sio.connected:
            return

        waiting_for_on_connect = asyncio.get_running_loop().create_future()
        self._waiting_for_on_connect = waiting_for_on_connect
        try:
            # sio.connect is raising a NotImplementedError, likely due to Windows and/or asyncqt?
            engineio.asyncio_client.async_signal_handler_set = True

            self.connection_state = ConnectionState.Connecting
            # self.logger.info(f"connect_to_server: sleeping")
            # await asyncio.sleep(1)
            # self.logger.info(f"connect_to_server: sleep over")

            self.logger.info(f"connecting to {self.configuration['server_address']}")
            self._connect_error = None
            await self.sio.connect(self.configuration["server_address"],
                                   socketio_path=self.configuration["socketio_path"],
                                   transports=['websocket'],
                                   headers={"X-Randovania-Version": randovania.VERSION})
            await waiting_for_on_connect
            self.logger.info(f"connected")

        except (socketio.exceptions.ConnectionError, aiohttp.client_exceptions.ContentTypeError) as e:
            self.logger.info(f"failed with {e} - {type(e)}")
            if self._connect_error is None:
                if isinstance(e, aiohttp.client_exceptions.ContentTypeError):
                    message = e.message
                else:
                    message = str(e)
                await self.on_connect_error(message)
            error = self._connect_error
            await self.sio.disconnect()
            raise UnableToConnect(error)

    def notify_on_connect(self, error_message: Optional[Exception]):
        if self._waiting_for_on_connect is not None:
            if error_message is None:
                self._waiting_for_on_connect.set_result(None)
            else:
                self._waiting_for_on_connect.set_exception(error_message)
            self._waiting_for_on_connect = None

    def connect_to_server(self) -> asyncio.Task:
        if self._connect_task is None:
            self._connect_task = asyncio.create_task(self._internal_connect_to_server())
            self._connect_task.add_done_callback(lambda _: setattr(self, "_connect_task", None))

        return self._connect_task

    async def disconnect_from_server(self):
        self.logger.debug(f"will disconnect")
        await self.sio.disconnect()
        self.logger.debug(f"disconnected")

    async def _restore_session(self):
        persisted_session = await self.read_persisted_session()
        if persisted_session is not None:
            if self._current_game_session is not None:
                session_id = self._current_game_session.id
            else:
                session_id = None
            try:
                self.connection_state = ConnectionState.ConnectedRestoringSession
                self.logger.debug(f"session restoring session, with id {session_id}")
                await self.on_user_session_updated(await self._emit_with_result("restore_user_session",
                                                                                (persisted_session, session_id)))

                if self._current_game_session is not None:
                    await self.on_game_session_updated(
                        await self._emit_with_result("game_session_request_update", self._current_game_session.id))
                    await self.on_game_update_notification({})

                self.logger.info(f"session restored successful")
                self.connection_state = ConnectionState.Connected
            except InvalidSession:
                self.logger.info(f"invalid session, deleting")
                self.connection_state = ConnectionState.ConnectedNotLogged
                self.session_data_path.unlink()
        else:
            self.logger.info(f"no session to restore")
            self.connection_state = ConnectionState.ConnectedNotLogged

    async def on_connect(self):
        try:
            self._restore_session_task = asyncio.create_task(self._restore_session())
            self._restore_session_task.add_done_callback(lambda _: setattr(self, "_restore_session_task", None))
            await self._restore_session_task
        finally:
            self.notify_on_connect(None)

    async def on_connect_error(self, error_message: str):
        if isinstance(error_message, dict) and "message" in error_message:
            error_message = error_message["message"]

        try:
            self._connect_error = error_message
            self.logger.warning(error_message)
            self.connection_state = ConnectionState.Disconnected
            if self._restore_session_task is not None:
                self._restore_session_task.cancel()
        finally:
            self.notify_on_connect(socketio.exceptions.ConnectionError(error_message))

    async def on_disconnect(self):
        self.logger.info(f"on_disconnect")
        self.connection_state = ConnectionState.Disconnected
        if self._restore_session_task is not None:
            self._restore_session_task.cancel()

    async def on_user_session_updated(self, new_session: dict):
        self._current_user = User.from_json(new_session["user"])
        if self.connection_state in (ConnectionState.ConnectedRestoringSession, ConnectionState.ConnectedNotLogged):
            self.connection_state = ConnectionState.Connected

        self.logger.info(f"{self._current_user.name}, state: {self.connection_state}")

        encoded_session_data = base64.b85decode(new_session["encoded_session_b85"])
        self.server_data_path.mkdir(exist_ok=True, parents=True)
        async with aiofiles.open(self.session_data_path, "wb") as open_file:
            await open_file.write(encoded_session_data)

    async def on_game_session_updated(self, data):
        self._current_game_session = GameSessionEntry.from_json(data)
        self.logger.debug(f"{self._current_game_session.id}")

    async def on_game_update_notification(self, details):
        pass

    async def _emit_with_result(self, event, data=None, namespace=None):
        if self.connection_state.is_disconnected:
            self.logger.debug(f"{event}, urgent connect start")
            await self.connect_to_server()
            self.logger.debug(f"{event}, urgent connect finished")

        self.logger.debug(f"{event}, getting lock")
        async with self._call_lock:
            self.logger.debug(f"{event}, will call")
            result = await self.sio.call(event, data, namespace=namespace, timeout=30)

        if result is None:
            return None

        possible_error = decode_error(result)
        if possible_error is None:
            return result["result"]
        else:
            raise possible_error

    async def game_session_collect_locations(self, locations: Tuple[int, ...]):
        await self._emit_with_result("game_session_collect_locations",
                                     (self._current_game_session.id, locations))

    async def game_session_request_pickups(self) -> List[Tuple[str, bytes]]:
        data = await self._emit_with_result("game_session_request_pickups", self._current_game_session.id)
        return [
            (item["provider_name"], base64.b85decode(item["pickup"]))
            for item in data
            if item is not None
        ]

    async def get_game_session_list(self) -> List[GameSessionListEntry]:
        return [
            GameSessionListEntry.from_json(item)
            for item in await self._emit_with_result("list_game_sessions")
        ]

    async def create_new_session(self, session_name: str) -> GameSessionEntry:
        result = await self._emit_with_result("create_game_session", session_name)
        self._current_game_session = GameSessionEntry.from_json(result)
        return self._current_game_session

    async def join_game_session(self, session: GameSessionListEntry, password: Optional[str]):
        result = await self._emit_with_result("join_game_session", (session.id, password))
        self._current_game_session = GameSessionEntry.from_json(result)

    async def leave_game_session(self, permanent: bool):
        if permanent:
            await self.session_admin_player(self._current_user.id, SessionAdminUserAction.KICK, None)
        await self._emit_with_result("disconnect_game_session", self._current_game_session.id)
        self._current_game_session = None
        self._last_self_update = None

    async def session_admin_global(self, action: SessionAdminGlobalAction, arg):
        return await self._emit_with_result("game_session_admin_session",
                                            (self._current_game_session.id, action.value, arg))

    async def session_admin_player(self, user_id: int, action: SessionAdminUserAction, arg):
        return await self._emit_with_result("game_session_admin_player",
                                            (self._current_game_session.id, user_id, action.value, arg))

    async def session_self_update(self,
                                  inventory: Dict[ItemResourceInfo, InventoryItem],
                                  state: GameConnectionStatus, backend: GameBackendChoice):

        inventory_json = json.dumps([
            {"index": resource.index, "amount": item.amount, "capacity": item.capacity}
            for resource, item in inventory.items()
        ])
        state_string = f"{state.pretty_text} ({backend.pretty_text})"

        if self._last_self_update != state_string:
            await self._emit_with_result("game_session_self_update",
                                         (self._current_game_session.id, inventory_json, state_string))
            self._last_self_update = state_string

    @property
    def current_user(self) -> Optional[User]:
        return self._current_user

    @property
    def current_game_session(self) -> Optional[GameSessionEntry]:
        return self._current_game_session



from typing import Tuple

from PySide2 import QtWidgets
from PySide2.QtCore import Signal
from PySide2.QtWidgets import QDialog, QWidget

from randovania.game_description.item.major_item import MajorItem
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.gui.generated.item_configuration_popup_ui import Ui_ItemConfigurationPopup
from randovania.gui.lib.common_qt_lib import set_default_window_icon
from randovania.layout.major_item_state import MajorItemState


class ItemConfigurationWidget(QDialog, Ui_ItemConfigurationPopup):
    Changed = Signal()

    def __init__(self, parent: QWidget, item: MajorItem, starting_state: MajorItemState,
                 resources_database: ResourceDatabase):
        super().__init__(parent)
        self.setupUi(self)
        set_default_window_icon(self)
        self._item = item

        self.setWindowTitle(f"Item: {item.name}")
        self.item_name_label.setText(item.name)

        # connect
        self.excluded_radio.toggled.connect(self._on_select_excluded)
        self.vanilla_radio.toggled.connect(self._on_select_vanilla)
        self.starting_radio.toggled.connect(self._on_select_starting)
        self.shuffled_radio.toggled.connect(self._on_select_shuffled)
        self.shuffled_spinbox.valueChanged.connect(self._on_shuffled_value)
        self.provided_ammo_spinbox.valueChanged.connect(self._on_shuffled_value)

        # Update
        self.vanilla_radio.setEnabled(item.original_index is not None)
        self.shuffled_radio.setEnabled(item.model_index is not None)

        if not self.vanilla_radio.isEnabled():
            self.vanilla_radio.setToolTip(
                "This item does not exist in the original game, so there's no vanilla location.")

        if not self.shuffled_radio.isEnabled():
            self.shuffled_radio.setToolTip(
                "There's currently no available model and/or logic for this item to be shuffled.")

        # At least one radio should be selected
        for radio in (self.shuffled_radio, self.starting_radio, self.vanilla_radio):
            if radio.isEnabled():
                radio.setChecked(True)
                break

        if item.ammo_index:
            self.provided_ammo_label.setText(
                "<html><head/><body><p>Provided Ammo ({})</p></body></html>".format(
                    " and ".join(
                        resources_database.get_item(ammo_index).long_name
                        for ammo_index in item.ammo_index
                    )
                )
            )
        else:
            self.provided_ammo_label.hide()
            self.provided_ammo_spinbox.hide()

        if self._item.required:
            self.item_name_label.setToolTip(
                "This item is necessary for the game to function properly and can't be removed.")
            self.setEnabled(False)
            self.state = self._create_state(num_included_in_starting_items=1)
        else:
            if self._item.warning is not None:
                self.item_name_label.setToolTip(self._item.warning)
            self.state = starting_state

    @property
    def item(self):
        return self._item

    @property
    def state(self) -> MajorItemState:
        if not self.excluded_radio.isChecked():
            return MajorItemState(
                include_copy_in_original_location=self.vanilla_radio.isChecked(),
                num_shuffled_pickups=self.shuffled_spinbox.value() if self.shuffled_radio.isChecked() else 0,
                num_included_in_starting_items=1 if self.starting_radio.isChecked() else 0,
                included_ammo=self.included_ammo,
            )
        else:
            return MajorItemState(
                include_copy_in_original_location=False,
                num_shuffled_pickups=0,
                num_included_in_starting_items=0,
                included_ammo=self.included_ammo,
            )

    @state.setter
    def state(self, value: MajorItemState):
        self._update_for_state(value)

    def _update_for_state(self, state):
        self.shuffled_spinbox.setEnabled(False)

        if state.included_ammo:
            self.provided_ammo_spinbox.setValue(state.included_ammo[0])

        if state.include_copy_in_original_location and self.vanilla_radio.isEnabled():
            self.vanilla_radio.setChecked(True)

        elif state.num_shuffled_pickups > 0 and self.shuffled_radio.isEnabled():
            self.shuffled_radio.setChecked(True)
            self.shuffled_spinbox.setEnabled(True)
            self.shuffled_spinbox.setValue(state.num_shuffled_pickups)

        elif state.num_included_in_starting_items > 0:
            self.starting_radio.setChecked(True)

        else:
            self.excluded_radio.setChecked(True)

        self.Changed.emit()

    def button_box_accepted(self):
        self.accept()

    def button_box_rejected(self):
        self.reject()

    def _on_included_box_toggle(self, enabled: bool):
        self._update_for_state(self.state)

    def _on_select_excluded(self, value: bool):
        self._update_for_state(self.state)

    def _on_select_vanilla(self, value: bool):
        self._update_for_state(self.state)

    def _on_select_starting(self, value: bool):
        self._update_for_state(self.state)

    def _on_select_shuffled(self, value: bool):
        self._update_for_state(self.state)

    def _on_shuffled_value(self, value: int):
        self._update_for_state(self.state)

    def _create_state(self,
                      *,
                      include_copy_in_original_location=False,
                      num_shuffled_pickups=0,
                      num_included_in_starting_items=0,
                      ):
        return MajorItemState(
            include_copy_in_original_location=include_copy_in_original_location,
            num_shuffled_pickups=num_shuffled_pickups,
            num_included_in_starting_items=num_included_in_starting_items,
            included_ammo=self.included_ammo,
        )

    @property
    def included_ammo(self) -> Tuple[int, ...]:
        return tuple(
            self.provided_ammo_spinbox.value()
            for _ in self._item.ammo_index
        )



import copy
from distutils.version import StrictVersion

import pytest

from randovania.interface_common import update_checker
from randovania.interface_common.update_checker import VersionDescription

_CUSTOM_CHANGE_LOGS = {
    4: "\n- *Major* - Foo\n",
    5: "\n- *Major* - Foo\n- Something\n",
    6: "\n- Bar\n",
}
_CUSTOM_EXPECTED_LOG = copy.copy(_CUSTOM_CHANGE_LOGS)
_CUSTOM_EXPECTED_LOG[5] = "\n- *Major* - Foo\n\n---\nFor more details, check the Change Log tab."


@pytest.mark.parametrize(["current_version", "last_changelog_version", "expected_display"], [
    (5, 5, 10),
    (5, 4, 10),
    (5, 3, 10),
    (10, 3, None),

])
def test_versions_to_display_for_releases(current_version, last_changelog_version, expected_display,
                                          ):
    # Setup
    releases = [
        {"tag_name": "v0.{}.0".format(i),
         "body": "Changelog v0.{}.0{}".format(i, _CUSTOM_CHANGE_LOGS.get(i, "")),
         "html_url": "url"}
        for i in reversed(range(1, 11))
    ]

    # Run
    all_change_logs, change_logs, version_to_display = update_checker.versions_to_display_for_releases(
        StrictVersion("0.{}.0".format(current_version)),
        StrictVersion("0.{}.0".format(last_changelog_version)),
        releases)

    # Assert
    if expected_display is None:
        assert version_to_display is None
    else:
        assert version_to_display == VersionDescription("v0.{}.0".format(expected_display),
                                                        "Changelog v0.{}.0".format(expected_display),
                                                        "url")
    assert change_logs == [
        "## v0.{0}.0\n\nChangelog v0.{0}.0{1}".format(i, _CUSTOM_EXPECTED_LOG.get(i, ""))
        for i in reversed(range(last_changelog_version + 1, current_version + 1))
    ]
    assert all_change_logs == [
        "## v0.{0}.0\n\nChangelog v0.{0}.0{1}".format(i, _CUSTOM_CHANGE_LOGS.get(i, ""))
        for i in reversed(range(1, current_version + 1))
    ]



import collections
import dataclasses
import functools
from functools import partial
from typing import Dict, Tuple, List

from PySide2 import QtWidgets
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QLabel, QGroupBox, QGridLayout, QSizePolicy, QSpinBox, \
    QCheckBox

from randovania.game_description import default_database
from randovania.game_description.item.ammo import Ammo
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.item.item_database import ItemDatabase
from randovania.game_description.item.major_item import MajorItem
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_type import ResourceType
from randovania.games.game import RandovaniaGame
from randovania.generator.item_pool import pool_creator
from randovania.generator.item_pool.ammo import items_for_ammo
from randovania.gui.generated.preset_item_pool_ui import Ui_PresetItemPool
from randovania.gui.lib import common_qt_lib
from randovania.gui.preset_settings.item_configuration_widget import ItemConfigurationWidget
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.gui.preset_settings.progressive_item_widget import ProgressiveItemWidget
from randovania.gui.preset_settings.split_ammo_widget import SplitAmmoWidget, AmmoPickupWidgets
from randovania.interface_common.enum_lib import iterate_enum
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.ammo_state import AmmoState
from randovania.layout.major_item_state import ENERGY_TANK_MAXIMUM_COUNT, DEFAULT_MAXIMUM_SHUFFLED, MajorItemState
from randovania.layout.preset import Preset
from randovania.resolver.exceptions import InvalidConfiguration

_EXPECTED_COUNT_TEXT_TEMPLATE = ("Each expansion will provide, on average, {per_expansion}, for a total of {total}."
                                 "\n{from_items} will be provided from major items.")
_EXPECTED_COUNT_TEXT_TEMPLATE_EXACT = ("Each expansion will provide exactly {per_expansion}, for a total of {total}."
                                       "\n{from_items} will be provided from major items.")


class PresetItemPool(PresetTab, Ui_PresetItemPool):
    _boxes_for_category: Dict[ItemCategory, Tuple[QGroupBox, QGridLayout, Dict[MajorItem, ItemConfigurationWidget]]]
    _default_items: Dict[ItemCategory, QtWidgets.QComboBox]

    _ammo_maximum_spinboxes: Dict[int, List[QSpinBox]]
    _ammo_pickup_widgets: Dict[Ammo, AmmoPickupWidgets]
    _progressive_widgets: List[ProgressiveItemWidget]
    _split_ammo_widgets: List[SplitAmmoWidget]

    def __init__(self, editor: PresetEditor):
        super().__init__()
        self.setupUi(self)

        self._editor = editor
        size_policy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        self.item_pool_layout.setAlignment(Qt.AlignTop)

        # Relevant Items
        self.game = editor.game
        self.game_description = default_database.game_description_for(self.game)
        item_database = default_database.item_database_for_game(self.game)

        self._energy_tank_item = item_database.major_items["Energy Tank"]

        self._register_random_starting_events()
        self._create_categories_boxes(item_database, size_policy)
        self._create_customizable_default_items(item_database)
        self._create_progressive_widgets(item_database)
        self._create_major_item_boxes(item_database, self.game_description.resource_database)
        self._create_energy_tank_box()
        self._create_split_ammo_widgets(item_database)
        self._create_ammo_pickup_boxes(size_policy, item_database)

    @property
    def uses_patches_tab(self) -> bool:
        return False

    def on_preset_changed(self, preset: Preset):
        # Item alternatives
        layout = preset.configuration
        major_configuration = layout.major_items_configuration

        for progressive_widget in self._progressive_widgets:
            progressive_widget.on_preset_changed(
                preset,
                self._boxes_for_category[progressive_widget.progressive_item.item_category][2],
            )

        for split_ammo in self._split_ammo_widgets:
            split_ammo.on_preset_changed(preset, self._ammo_pickup_widgets)

        # Random Starting Items
        self.minimum_starting_spinbox.setValue(major_configuration.minimum_random_starting_items)
        self.maximum_starting_spinbox.setValue(major_configuration.maximum_random_starting_items)

        # Default Items
        for category, default_item in major_configuration.default_items.items():
            combo = self._default_items[category]
            combo.setCurrentIndex(combo.findData(default_item))

            for item, widget in self._boxes_for_category[category][2].items():
                widget.setEnabled(default_item != item)

        # Major Items
        for _, _, elements in self._boxes_for_category.values():
            for major_item, widget in elements.items():
                widget.state = major_configuration.items_state[major_item]

        # Energy Tank
        energy_tank_state = major_configuration.items_state[self._energy_tank_item]
        self.energy_tank_starting_spinbox.setValue(energy_tank_state.num_included_in_starting_items)
        self.energy_tank_shuffled_spinbox.setValue(energy_tank_state.num_shuffled_pickups)

        # Ammo
        ammo_provided = major_configuration.calculate_provided_ammo()
        ammo_configuration = layout.ammo_configuration

        for ammo_item, maximum in ammo_configuration.maximum_ammo.items():
            for spinbox in self._ammo_maximum_spinboxes[ammo_item]:
                spinbox.setValue(maximum)

        previous_pickup_for_item = {}
        resource_database = self.game_description.resource_database

        item_for_index: Dict[int, ItemResourceInfo] = {
            ammo_index: resource_database.get_item(ammo_index)
            for ammo_index in ammo_provided.keys()
        }

        for ammo, state in ammo_configuration.items_state.items():
            widgets = self._ammo_pickup_widgets[ammo]
            widgets.pickup_spinbox.setValue(state.pickup_count)

            if widgets.require_major_item_check is not None:
                widgets.require_major_item_check.setChecked(state.requires_major_item)

            try:
                if state.pickup_count == 0:
                    widgets.expected_count.setText("No expansions will be created.")
                    continue

                ammo_per_pickup = items_for_ammo(ammo, state, ammo_provided,
                                                 previous_pickup_for_item,
                                                 ammo_configuration.maximum_ammo)

                totals = functools.reduce(lambda a, b: [x + y for x, y in zip(a, b)],
                                          ammo_per_pickup,
                                          [0 for _ in ammo.items])

                if {total % state.pickup_count for total in totals} == {0}:
                    count_text_template = _EXPECTED_COUNT_TEXT_TEMPLATE_EXACT
                else:
                    count_text_template = _EXPECTED_COUNT_TEXT_TEMPLATE

                widgets.expected_count.setText(
                    count_text_template.format(
                        per_expansion=" and ".join(
                            "{:.3g} {}".format(
                                total / state.pickup_count,
                                item_for_index[ammo_index].long_name
                            )
                            for ammo_index, total in zip(ammo.items, totals)
                        ),
                        total=" and ".join(
                            "{} {}".format(
                                total,
                                item_for_index[ammo_index].long_name
                            )
                            for ammo_index, total in zip(ammo.items, totals)
                        ),
                        from_items=" and ".join(
                            "{} {}".format(
                                ammo_provided[ammo_index],
                                item_for_index[ammo_index].long_name
                            )
                            for ammo_index in ammo.items
                        ),
                    )
                )

            except InvalidConfiguration as invalid_config:
                widgets.expected_count.setText(str(invalid_config))

        # Item pool count
        try:
            pool_pickup = pool_creator.calculate_pool_results(layout, resource_database).pickups
            min_starting_items = layout.major_items_configuration.minimum_random_starting_items
            maximum_size = self.game_description.world_list.num_pickup_nodes + min_starting_items
            self.item_pool_count_label.setText(
                "Items in pool: {}/{}".format(len(pool_pickup), maximum_size)
            )
            common_qt_lib.set_error_border_stylesheet(self.item_pool_count_label, len(pool_pickup) > maximum_size)

        except InvalidConfiguration as invalid_config:
            self.item_pool_count_label.setText("Invalid Configuration: {}".format(invalid_config))
            common_qt_lib.set_error_border_stylesheet(self.item_pool_count_label, True)

    # Item Alternatives

    def _persist_bool_layout_field(self, field_name: str):
        def bound(value: int):
            with self._editor as options:
                options.set_configuration_field(field_name, bool(value))

        return bound

    def _persist_bool_major_configuration_field(self, field_name: str):
        def bound(value: int):
            with self._editor as options:
                kwargs = {field_name: bool(value)}
                options.major_items_configuration = dataclasses.replace(options.major_items_configuration, **kwargs)

        return bound

    # Random Starting

    def _register_random_starting_events(self):
        self.minimum_starting_spinbox.valueChanged.connect(self._on_update_minimum_starting)
        self.maximum_starting_spinbox.valueChanged.connect(self._on_update_maximum_starting)

    def _on_update_minimum_starting(self, value: int):
        with self._editor as options:
            options.major_items_configuration = dataclasses.replace(options.major_items_configuration,
                                                                    minimum_random_starting_items=value)

    def _on_update_maximum_starting(self, value: int):
        with self._editor as options:
            options.major_items_configuration = dataclasses.replace(options.major_items_configuration,
                                                                    maximum_random_starting_items=value)

    # Major Items
    def _create_progressive_widgets(self, item_database: ItemDatabase):
        self._progressive_widgets = []

        all_progressive = []

        if self.game == RandovaniaGame.PRIME2:
            all_progressive.append((
                "Progressive Suit",
                ("Dark Suit", "Light Suit"),
            ))
            all_progressive.append((
                "Progressive Grapple",
                ("Grapple Beam", "Screw Attack"),
            ))
        elif self.game == RandovaniaGame.PRIME3:
            all_progressive.append((
                "Progressive Missile",
                ("Ice Missile", "Seeker Missile"),
            ))
            all_progressive.append((
                "Progressive Beam",
                ("Plasma Beam", "Nova Beam"),
            ))

        for (progressive_item_name, non_progressive_items) in all_progressive:
            progressive_item = item_database.major_items[progressive_item_name]
            parent, layout, _ = self._boxes_for_category[progressive_item.item_category]

            widget = ProgressiveItemWidget(
                parent, self._editor,
                progressive_item=progressive_item,
                non_progressive_items=[item_database.major_items[it] for it in non_progressive_items],
            )
            widget.setText("Use progressive {}".format(" → ".join(non_progressive_items)))
            self._progressive_widgets.append(widget)
            layout.addWidget(widget)

    def _create_split_ammo_widgets(self, item_database: ItemDatabase):
        parent, layout, _ = self._boxes_for_category[ItemCategory.BEAM]

        self._split_ammo_widgets = []

        if self.game == RandovaniaGame.PRIME2:
            beam_ammo = SplitAmmoWidget(
                parent, self._editor,
                unified_ammo=item_database.ammo["Beam Ammo Expansion"],
                split_ammo=[
                    item_database.ammo["Dark Ammo Expansion"],
                    item_database.ammo["Light Ammo Expansion"],
                ],
            )
            beam_ammo.setText("Split Beam Ammo Expansions")
            self._split_ammo_widgets.append(beam_ammo)

        if self._split_ammo_widgets:
            line = QtWidgets.QFrame(parent)
            line.setFrameShape(QtWidgets.QFrame.HLine)
            line.setFrameShadow(QtWidgets.QFrame.Sunken)
            layout.addWidget(line, layout.rowCount(), 0, 1, -1)

        for widget in self._split_ammo_widgets:
            layout.addWidget(widget, layout.rowCount(), 0, 1, -1)

    def _create_categories_boxes(self, item_database: ItemDatabase, size_policy):
        self._boxes_for_category = {}

        categories = set()
        for major_item in item_database.major_items.values():
            if not major_item.required:
                categories.add(major_item.item_category)

        all_categories = list(iterate_enum(ItemCategory))
        for major_item_category in sorted(categories, key=lambda it: all_categories.index(it)):
            category_box = QGroupBox(self.scroll_area_contents)
            category_box.setTitle(major_item_category.long_name)
            category_box.setSizePolicy(size_policy)
            category_box.setObjectName(f"category_box {major_item_category}")

            category_layout = QGridLayout(category_box)
            category_layout.setObjectName(f"category_layout {major_item_category}")

            self.item_pool_layout.addWidget(category_box)
            self._boxes_for_category[major_item_category] = category_box, category_layout, {}

    def _create_customizable_default_items(self, item_database: ItemDatabase):
        self._default_items = {}

        for category, possibilities in item_database.default_items.items():
            parent, layout, _ = self._boxes_for_category[category]

            label = QtWidgets.QLabel(parent)
            label.setText(f"Default {category.long_name}")
            layout.addWidget(label, 0, 0)

            combo = QtWidgets.QComboBox(parent)
            for item in possibilities:
                combo.addItem(item.name, item)
            combo.currentIndexChanged.connect(partial(self._on_default_item_updated, category, combo))
            layout.addWidget(combo, 0, 1)

            self._default_items[category] = combo

    def _on_default_item_updated(self, category: ItemCategory, combo: QtWidgets.QComboBox, _):
        with self._editor as editor:
            new_config = editor.major_items_configuration
            new_config = new_config.replace_default_item(category, combo.currentData())
            new_config = new_config.replace_state_for_item(combo.currentData(),
                                                           MajorItemState(num_included_in_starting_items=1))
            editor.major_items_configuration = new_config

    def _create_major_item_boxes(self, item_database: ItemDatabase, resource_database: ResourceDatabase):
        for major_item in item_database.major_items.values():
            if major_item.required or major_item.item_category == ItemCategory.ENERGY_TANK:
                continue

            category_box, category_layout, elements = self._boxes_for_category[major_item.item_category]
            widget = ItemConfigurationWidget(None, major_item, MajorItemState(), resource_database)
            widget.Changed.connect(partial(self._on_major_item_updated, widget))

            row = category_layout.rowCount()
            category_layout.addWidget(widget, row, 0, 1, -1)

            elements[major_item] = widget

    def _on_major_item_updated(self, item_widget: ItemConfigurationWidget):
        with self._editor as editor:
            editor.major_items_configuration = editor.major_items_configuration.replace_state_for_item(
                item_widget.item, item_widget.state
            )

    # Energy Tank

    def _create_energy_tank_box(self):
        category_box, category_layout, _ = self._boxes_for_category[ItemCategory.ENERGY_TANK]

        starting_label = QLabel(category_box)
        starting_label.setText("Starting Quantity")

        shuffled_label = QLabel(category_box)
        shuffled_label.setText("Shuffled Quantity")

        self.energy_tank_starting_spinbox = QSpinBox(category_box)
        self.energy_tank_starting_spinbox.setMaximum(ENERGY_TANK_MAXIMUM_COUNT)
        self.energy_tank_starting_spinbox.valueChanged.connect(self._on_update_starting_energy_tank)
        self.energy_tank_shuffled_spinbox = QSpinBox(category_box)
        self.energy_tank_shuffled_spinbox.setMaximum(DEFAULT_MAXIMUM_SHUFFLED[-1])
        self.energy_tank_shuffled_spinbox.valueChanged.connect(self._on_update_shuffled_energy_tank)

        category_layout.addWidget(starting_label, 0, 0)
        category_layout.addWidget(self.energy_tank_starting_spinbox, 0, 1)
        category_layout.addWidget(shuffled_label, 1, 0)
        category_layout.addWidget(self.energy_tank_shuffled_spinbox, 1, 1)

    def _on_update_starting_energy_tank(self, value: int):
        with self._editor as options:
            major_configuration = options.major_items_configuration
            options.major_items_configuration = major_configuration.replace_state_for_item(
                self._energy_tank_item,
                dataclasses.replace(major_configuration.items_state[self._energy_tank_item],
                                    num_included_in_starting_items=value)
            )

    def _on_update_shuffled_energy_tank(self, value: int):
        with self._editor as options:
            major_configuration = options.major_items_configuration
            options.major_items_configuration = major_configuration.replace_state_for_item(
                self._energy_tank_item,
                dataclasses.replace(major_configuration.items_state[self._energy_tank_item],
                                    num_shuffled_pickups=value)
            )

    # Ammo

    def _create_ammo_pickup_boxes(self, size_policy, item_database: ItemDatabase):
        """
        Creates the GroupBox with SpinBoxes for selecting the pickup count of all the ammo
        :param item_database:
        :return:
        """

        self._ammo_maximum_spinboxes = collections.defaultdict(list)
        self._ammo_pickup_widgets = {}

        resource_database = default_database.resource_database_for(self.game)
        broad_to_category = {
            ItemCategory.BEAM_RELATED: ItemCategory.BEAM,
            ItemCategory.MORPH_BALL_RELATED: ItemCategory.MORPH_BALL,
            ItemCategory.MISSILE_RELATED: ItemCategory.MISSILE,
        }

        for ammo in item_database.ammo.values():
            category_box, category_layout, _ = self._boxes_for_category[broad_to_category[ammo.broad_category]]

            pickup_box = QGroupBox(category_box)
            pickup_box.setSizePolicy(size_policy)
            pickup_box.setTitle(ammo.name + "s")
            layout = QGridLayout(pickup_box)
            layout.setObjectName(f"{ammo.name} Box Layout")
            current_row = 0

            for ammo_item in ammo.items:
                item = resource_database.get_by_type_and_index(ResourceType.ITEM, ammo_item)

                target_count_label = QLabel(pickup_box)
                target_count_label.setText(f"{item.long_name} Target" if len(ammo.items) > 1 else "Target count")

                maximum_spinbox = QSpinBox(pickup_box)
                maximum_spinbox.setMaximum(ammo.maximum)
                maximum_spinbox.valueChanged.connect(partial(self._on_update_ammo_maximum_spinbox, ammo_item))
                self._ammo_maximum_spinboxes[ammo_item].append(maximum_spinbox)

                layout.addWidget(target_count_label, current_row, 0)
                layout.addWidget(maximum_spinbox, current_row, 1)
                current_row += 1

            count_label = QLabel(pickup_box)
            count_label.setText("Pickup Count")
            count_label.setToolTip("How many instances of this expansion should be placed.")

            pickup_spinbox = QSpinBox(pickup_box)
            pickup_spinbox.setMaximum(AmmoState.maximum_pickup_count())
            pickup_spinbox.valueChanged.connect(partial(self._on_update_ammo_pickup_spinbox, ammo))

            layout.addWidget(count_label, current_row, 0)
            layout.addWidget(pickup_spinbox, current_row, 1)
            current_row += 1

            if ammo.temporaries:
                require_major_item_check = QCheckBox(pickup_box)
                require_major_item_check.setText("Requires the major item to work?")
                require_major_item_check.stateChanged.connect(partial(self._on_update_ammo_require_major_item, ammo))
                layout.addWidget(require_major_item_check, current_row, 0, 1, 2)
                current_row += 1
            else:
                require_major_item_check = None

            expected_count = QLabel(pickup_box)
            expected_count.setWordWrap(True)
            expected_count.setText(_EXPECTED_COUNT_TEXT_TEMPLATE)
            expected_count.setToolTip("Some expansions may provide 1 extra, even with no variance, if the total count "
                                      "is not divisible by the pickup count.")
            layout.addWidget(expected_count, current_row, 0, 1, 2)
            current_row += 1

            self._ammo_pickup_widgets[ammo] = AmmoPickupWidgets(pickup_spinbox, expected_count,
                                                                pickup_box, require_major_item_check)
            category_layout.addWidget(pickup_box)

    def _on_update_ammo_maximum_spinbox(self, ammo_int: int, value: int):
        with self._editor as options:
            options.ammo_configuration = options.ammo_configuration.replace_maximum_for_item(
                ammo_int, value
            )

    def _on_update_ammo_pickup_spinbox(self, ammo: Ammo, value: int):
        with self._editor as options:
            ammo_configuration = options.ammo_configuration
            options.ammo_configuration = ammo_configuration.replace_state_for_ammo(
                ammo,
                dataclasses.replace(ammo_configuration.items_state[ammo], pickup_count=value)
            )

    def _on_update_ammo_require_major_item(self, ammo: Ammo, value: int):
        with self._editor as options:
            ammo_configuration = options.ammo_configuration
            options.ammo_configuration = ammo_configuration.replace_state_for_ammo(
                ammo,
                dataclasses.replace(ammo_configuration.items_state[ammo], requires_major_item=bool(value))
            )



import functools
import functools
import json
import os
import platform
import subprocess
from functools import partial
from typing import Optional, List

import markdown
from PySide2 import QtCore, QtWidgets, QtGui
from PySide2.QtCore import QUrl, Signal, Qt
from PySide2.QtGui import QDesktopServices
from PySide2.QtWidgets import QMainWindow, QAction, QMessageBox, QDialog, QMenu, QInputDialog
from asyncqt import asyncSlot

from randovania import VERSION
from randovania.game_description import default_database
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.node import LogbookNode, LoreType
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo
from randovania.games.game import RandovaniaGame
from randovania.gui.data_editor import DataEditorWindow
from randovania.gui.dialog.login_prompt_dialog import LoginPromptDialog
from randovania.gui.dialog.permalink_dialog import PermalinkDialog
from randovania.gui.dialog.trick_details_popup import TrickDetailsPopup
from randovania.gui.game_session_window import GameSessionWindow
from randovania.gui.generate_seed_tab import GenerateSeedTab
from randovania.gui.generated.main_window_ui import Ui_MainWindow
from randovania.gui.lib import common_qt_lib, async_dialog, theme
from randovania.gui.lib.qt_network_client import handle_network_errors, QtNetworkClient
from randovania.gui.lib.trick_lib import used_tricks, difficulties_for_trick
from randovania.gui.lib.window_manager import WindowManager
from randovania.gui.online_game_list_window import GameSessionBrowserDialog
from randovania.gui.tracker_window import TrackerWindow, InvalidLayoutForTracker
from randovania.interface_common import github_releases_data, update_checker
from randovania.interface_common.enum_lib import iterate_enum
from randovania.interface_common.options import Options
from randovania.interface_common.preset_manager import PresetManager
from randovania.layout.echoes_configuration import EchoesConfiguration
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.trick_level import LayoutTrickLevel
from randovania.network_client.network_client import ConnectionState
from randovania.resolver import debug

_DISABLE_VALIDATION_WARNING = """
<html><head/><body>
<p>While it sometimes throws errors, the validation is what guarantees that your seed is completable.<br/>
Do <span style=" font-weight:600;">not</span> disable if you're uncomfortable with possibly unbeatable seeds.
</p><p align="center">Are you sure you want to disable validation?</p></body></html>
"""


class MainWindow(WindowManager, Ui_MainWindow):
    newer_version_signal = Signal(str, str)
    options_changed_signal = Signal()
    _is_preview_mode: bool = False

    menu_new_version: Optional[QAction] = None
    _current_version_url: Optional[str] = None
    _options: Options
    _data_visualizer: Optional[DataEditorWindow] = None
    _map_tracker: TrackerWindow
    _preset_manager: PresetManager
    game_session_window: Optional[GameSessionWindow] = None
    _login_window: Optional[QDialog] = None
    GameDetailsSignal = Signal(LayoutDescription)

    @property
    def _tab_widget(self):
        return self.main_tab_widget

    @property
    def preset_manager(self) -> PresetManager:
        return self._preset_manager

    @property
    def main_window(self) -> QMainWindow:
        return self

    @property
    def is_preview_mode(self) -> bool:
        return self._is_preview_mode

    def __init__(self, options: Options, preset_manager: PresetManager,
                 network_client: QtNetworkClient, preview: bool):
        super().__init__()
        self.setupUi(self)
        self.setWindowTitle("Randovania {}".format(VERSION))
        self._is_preview_mode = preview
        self.setAcceptDrops(True)
        common_qt_lib.set_default_window_icon(self)

        # Remove all hardcoded link color
        about_document: QtGui.QTextDocument = self.about_text_browser.document()
        about_document.setHtml(about_document.toHtml().replace("color:#0000ff;", ""))
        self.browse_racetime_label.setText(self.browse_racetime_label.text().replace("color:#0000ff;", ""))

        self.intro_label.setText(self.intro_label.text().format(version=VERSION))

        self._preset_manager = preset_manager
        self.network_client = network_client

        if preview:
            debug.set_level(2)

        # Signals
        self.newer_version_signal.connect(self.display_new_version)
        self.options_changed_signal.connect(self.on_options_changed)
        self.GameDetailsSignal.connect(self._open_game_details)

        self.intro_play_now_button.clicked.connect(lambda: self.welcome_tab_widget.setCurrentWidget(self.tab_play))
        self.open_faq_button.clicked.connect(self._open_faq)
        self.open_database_viewer_button.clicked.connect(partial(self._open_data_visualizer_for_game,
                                                                 RandovaniaGame.PRIME2))

        self.import_permalink_button.clicked.connect(self._import_permalink)
        self.import_game_file_button.clicked.connect(self._import_spoiler_log)
        self.browse_racetime_button.clicked.connect(self._browse_racetime)
        self.browse_sessions_button.clicked.connect(self._browse_for_game_session)
        self.host_new_game_button.clicked.connect(self._host_game_session)
        self.create_new_seed_button.clicked.connect(
            lambda: self.welcome_tab_widget.setCurrentWidget(self.tab_create_seed))

        # Menu Bar
        for action, game in ((self.menu_action_prime_1_data_visualizer, RandovaniaGame.PRIME1),
                             (self.menu_action_prime_2_data_visualizer, RandovaniaGame.PRIME2),
                             (self.menu_action_prime_3_data_visualizer, RandovaniaGame.PRIME3)):
            action.triggered.connect(partial(self._open_data_visualizer_for_game, game))

        for action, game in ((self.menu_action_edit_prime_1, RandovaniaGame.PRIME1),
                             (self.menu_action_edit_prime_2, RandovaniaGame.PRIME2),
                             (self.menu_action_edit_prime_3, RandovaniaGame.PRIME3)):
            action.triggered.connect(partial(self._open_data_editor_for_game, game))

        self.menu_action_item_tracker.triggered.connect(self._open_item_tracker)
        self.menu_action_map_tracker.triggered.connect(self._on_menu_action_map_tracker)
        self.menu_action_edit_existing_database.triggered.connect(self._open_data_editor_prompt)
        self.menu_action_validate_seed_after.triggered.connect(self._on_validate_seed_change)
        self.menu_action_timeout_generation_after_a_time_limit.triggered.connect(self._on_generate_time_limit_change)
        self.menu_action_dark_mode.triggered.connect(self._on_menu_action_dark_mode)
        self.menu_action_open_auto_tracker.triggered.connect(self._open_auto_tracker)
        self.menu_action_previously_generated_games.triggered.connect(self._on_menu_action_previously_generated_games)
        self.menu_action_layout_editor.triggered.connect(self._on_menu_action_layout_editor)
        self.action_login_window.triggered.connect(self._action_login_window)

        self.menu_prime_1_trick_details.aboutToShow.connect(self._create_trick_details_prime_1)
        self.menu_prime_2_trick_details.aboutToShow.connect(self._create_trick_details_prime_2)
        self.menu_prime_3_trick_details.aboutToShow.connect(self._create_trick_details_prime_3)

        self.generate_seed_tab = GenerateSeedTab(self, self, options)
        self.generate_seed_tab.setup_ui()

        # Setting this event only now, so all options changed trigger only once
        options.on_options_changed = self.options_changed_signal.emit
        self._options = options
        with options:
            self.on_options_changed()

        self.main_tab_widget.setCurrentIndex(0)

        # Update hints text
        self._update_hints_text()

    def closeEvent(self, event):
        self.generate_seed_tab.stop_background_process()
        super().closeEvent(event)

    # Generate Seed
    def _open_faq(self):
        self.main_tab_widget.setCurrentWidget(self.help_tab)
        self.help_tab_widget.setCurrentWidget(self.tab_faq)

    async def generate_seed_from_permalink(self, permalink):
        from randovania.interface_common.status_update_lib import ProgressUpdateCallable
        from randovania.gui.dialog.background_process_dialog import BackgroundProcessDialog

        def work(progress_update: ProgressUpdateCallable):
            from randovania.interface_common import simplified_patcher
            layout = simplified_patcher.generate_layout(progress_update=progress_update,
                                                        permalink=permalink,
                                                        options=self._options)
            progress_update(f"Success! (Seed hash: {layout.shareable_hash})", 1)
            return layout

        new_layout = await BackgroundProcessDialog.open_for_background_task(work, "Creating a game...")
        self.open_game_details(new_layout)

    @asyncSlot()
    async def _import_permalink(self):
        dialog = PermalinkDialog()
        result = await async_dialog.execute_dialog(dialog)
        if result == QDialog.Accepted:
            permalink = dialog.get_permalink_from_field()
            await self.generate_seed_from_permalink(permalink)

    def _import_spoiler_log(self):
        json_path = common_qt_lib.prompt_user_for_input_game_log(self)
        if json_path is not None:
            layout = LayoutDescription.from_file(json_path)
            self.open_game_details(layout)

    @asyncSlot()
    async def _browse_racetime(self):
        from randovania.gui.dialog.racetime_browser_dialog import RacetimeBrowserDialog
        dialog = RacetimeBrowserDialog()
        if not await dialog.refresh():
            return
        result = await async_dialog.execute_dialog(dialog)
        if result == QDialog.Accepted:
            await self.generate_seed_from_permalink(dialog.permalink)

    async def _game_session_active(self) -> bool:
        if self.game_session_window is None or self.game_session_window.has_closed:
            return False
        else:
            await async_dialog.message_box(
                self,
                QtWidgets.QMessageBox.Critical,
                "Game Session in progress",
                "There's already a game session window open. Please close it first.",
                QMessageBox.Ok
            )
            self.game_session_window.activateWindow()
            return True

    async def _ensure_logged_in(self) -> bool:
        network_client = self.network_client
        if network_client.connection_state == ConnectionState.Connected:
            return True

        if network_client.connection_state.is_disconnected:
            message_box = QtWidgets.QMessageBox(QtWidgets.QMessageBox.NoIcon, "Connecting",
                                                "Connecting to server...", QtWidgets.QMessageBox.Cancel,
                                                self)

            connecting = network_client.connect_to_server()
            message_box.rejected.connect(connecting.cancel)
            message_box.show()
            try:
                await connecting
            finally:
                message_box.close()

        if network_client.current_user is None:
            await async_dialog.execute_dialog(LoginPromptDialog(network_client))

        return network_client.current_user is not None

    @asyncSlot()
    @handle_network_errors
    async def _browse_for_game_session(self):
        if await self._game_session_active():
            return

        if not await self._ensure_logged_in():
            return

        network_client = self.network_client
        browser = GameSessionBrowserDialog(network_client)
        await browser.refresh()
        if await async_dialog.execute_dialog(browser) == browser.Accepted:
            self.game_session_window = await GameSessionWindow.create_and_update(
                network_client, common_qt_lib.get_game_connection(), self.preset_manager,
                self, self._options)
            if self.game_session_window is not None:
                self.game_session_window.show()

    @asyncSlot()
    @handle_network_errors
    async def _action_login_window(self):
        if self._login_window is not None:
            return self._login_window.show()

        self._login_window = LoginPromptDialog(self.network_client)
        try:
            await async_dialog.execute_dialog(self._login_window)
        finally:
            self._login_window = None

    @asyncSlot()
    @handle_network_errors
    async def _host_game_session(self):
        if await self._game_session_active():
            return

        if not await self._ensure_logged_in():
            return

        dialog = QInputDialog(self)
        dialog.setModal(True)
        dialog.setWindowTitle("Enter session name")
        dialog.setLabelText("Select a name for the session:")
        if await async_dialog.execute_dialog(dialog) != dialog.Accepted:
            return

        await self.network_client.create_new_session(dialog.textValue())
        self.game_session_window = await GameSessionWindow.create_and_update(self.network_client,
                                                                             common_qt_lib.get_game_connection(),
                                                                             self.preset_manager, self, self._options)
        if self.game_session_window is not None:
            self.game_session_window.show()

    def open_game_details(self, layout: LayoutDescription):
        self.GameDetailsSignal.emit(layout)

    def _open_game_details(self, layout: LayoutDescription):
        from randovania.gui.seed_details_window import SeedDetailsWindow
        details_window = SeedDetailsWindow(self, self._options)
        details_window.update_layout_description(layout)
        details_window.show()
        self.track_window(details_window)

    # Releases info
    async def request_new_data(self):
        await self._on_releases_data(await github_releases_data.get_releases())

    async def _on_releases_data(self, releases: Optional[List[dict]]):
        current_version = update_checker.strict_current_version()
        last_changelog = self._options.last_changelog_displayed

        all_change_logs, new_change_logs, version_to_display = update_checker.versions_to_display_for_releases(
            current_version, last_changelog, releases)

        if version_to_display is not None:
            self.display_new_version(version_to_display)

        if all_change_logs:
            changelog_tab = QtWidgets.QWidget()
            changelog_tab.setObjectName("changelog_tab")
            changelog_tab_layout = QtWidgets.QVBoxLayout(changelog_tab)
            changelog_tab_layout.setContentsMargins(0, 0, 0, 0)
            changelog_tab_layout.setObjectName("changelog_tab_layout")
            changelog_scroll_area = QtWidgets.QScrollArea(changelog_tab)
            changelog_scroll_area.setWidgetResizable(True)
            changelog_scroll_area.setObjectName("changelog_scroll_area")
            changelog_scroll_contents = QtWidgets.QWidget()
            changelog_scroll_contents.setGeometry(QtCore.QRect(0, 0, 489, 337))
            changelog_scroll_contents.setObjectName("changelog_scroll_contents")
            changelog_scroll_layout = QtWidgets.QVBoxLayout(changelog_scroll_contents)
            changelog_scroll_layout.setObjectName("changelog_scroll_layout")
            changelog_label = QtWidgets.QLabel(changelog_scroll_contents)
            changelog_label.setObjectName("changelog_label")
            changelog_label.setText(markdown.markdown("\n".join(all_change_logs)))
            changelog_label.setWordWrap(True)
            changelog_scroll_layout.addWidget(changelog_label)
            changelog_scroll_area.setWidget(changelog_scroll_contents)
            changelog_tab_layout.addWidget(changelog_scroll_area)
            self.help_tab_widget.addTab(changelog_tab, "Change Log")

        if new_change_logs:
            await async_dialog.message_box(self, QtWidgets.QMessageBox.Information,
                                           "What's new", markdown.markdown("\n".join(new_change_logs)))
            with self._options as options:
                options.last_changelog_displayed = current_version

    def display_new_version(self, version: update_checker.VersionDescription):
        if self.menu_new_version is None:
            self.menu_new_version = QAction("", self)
            self.menu_new_version.triggered.connect(self.open_version_link)
            self.menu_bar.addAction(self.menu_new_version)

        self.menu_new_version.setText("New version available: {}".format(version.tag_name))
        self._current_version_url = version.html_url

    def open_version_link(self):
        if self._current_version_url is None:
            raise RuntimeError("Called open_version_link, but _current_version_url is None")

        QDesktopServices.openUrl(QUrl(self._current_version_url))

    # Options
    def on_options_changed(self):
        self.menu_action_validate_seed_after.setChecked(self._options.advanced_validate_seed_after)
        self.menu_action_timeout_generation_after_a_time_limit.setChecked(
            self._options.advanced_timeout_during_generation)
        self.menu_action_dark_mode.setChecked(self._options.dark_mode)

        self.generate_seed_tab.on_options_changed(self._options)
        theme.set_dark_theme(self._options.dark_mode)

    # Menu Actions
    def _open_data_visualizer_for_game(self, game: RandovaniaGame):
        self.open_data_visualizer_at(None, None, game)

    def open_data_visualizer_at(self,
                                world_name: Optional[str],
                                area_name: Optional[str],
                                game: RandovaniaGame = RandovaniaGame.PRIME2,
                                ):
        self._data_visualizer = DataEditorWindow.open_internal_data(game, False)

        if world_name is not None:
            self._data_visualizer.focus_on_world(world_name)

        if area_name is not None:
            self._data_visualizer.focus_on_area(area_name)

        self._data_visualizer.show()

    def _open_data_editor_for_game(self, game: RandovaniaGame):
        self._data_editor = DataEditorWindow.open_internal_data(game, True)
        self._data_editor.show()

    def _open_data_editor_prompt(self):
        database_path = common_qt_lib.prompt_user_for_database_file(self)
        if database_path is None:
            return

        with database_path.open("r") as database_file:
            self._data_editor = DataEditorWindow(json.load(database_file), database_path, False, True)
            self._data_editor.show()

    @asyncSlot()
    async def _on_menu_action_map_tracker(self):
        dialog = QtWidgets.QInputDialog(self)
        dialog.setWindowTitle("Map Tracker")
        dialog.setLabelText("Select preset used for the tracker.")
        dialog.setComboBoxItems([preset.name for preset in self._preset_manager.all_presets])
        dialog.setTextValue(self._options.selected_preset_name)
        result = await async_dialog.execute_dialog(dialog)
        if result == QtWidgets.QDialog.Accepted:
            preset = self._preset_manager.preset_for_name(dialog.textValue())
            self.open_map_tracker(preset.get_preset().configuration)

    def open_map_tracker(self, configuration: EchoesConfiguration):
        try:
            self._map_tracker = TrackerWindow(self._options.tracker_files_path, configuration)
        except InvalidLayoutForTracker as e:
            QMessageBox.critical(
                self,
                "Unsupported configuration for Tracker",
                str(e)
            )
            return

        self._map_tracker.show()

    def _open_item_tracker(self):
        # Importing this at root level seems to crash linux tests :(
        from PySide2.QtWebEngineWidgets import QWebEngineView

        tracker_window = QMainWindow()
        tracker_window.setWindowTitle("Item Tracker")
        tracker_window.resize(370, 380)

        web_view = QWebEngineView(tracker_window)
        tracker_window.setCentralWidget(web_view)

        self.web_view = web_view

        def update_window_icon():
            tracker_window.setWindowIcon(web_view.icon())

        web_view.iconChanged.connect(update_window_icon)
        web_view.load(QUrl("https://spaghettitoastbook.github.io/echoes/tracker/"))

        tracker_window.show()
        self._item_tracker_window = tracker_window

    # Difficulties stuff

    def _exec_trick_details(self, popup: TrickDetailsPopup):
        self._trick_details_popup = popup
        self._trick_details_popup.setWindowModality(Qt.WindowModal)
        self._trick_details_popup.open()

    def _open_trick_details_popup(self, game, trick: TrickResourceInfo, level: LayoutTrickLevel):
        self._exec_trick_details(TrickDetailsPopup(self, self, game, trick, level))

    def _create_trick_details_prime_1(self):
        self.menu_prime_1_trick_details.aboutToShow.disconnect(self._create_trick_details_prime_1)
        self._setup_difficulties_menu(RandovaniaGame.PRIME1, self.menu_prime_1_trick_details)

    def _create_trick_details_prime_2(self):
        self.menu_prime_2_trick_details.aboutToShow.disconnect(self._create_trick_details_prime_2)
        self._setup_difficulties_menu(RandovaniaGame.PRIME2, self.menu_prime_2_trick_details)

    def _create_trick_details_prime_3(self):
        self.menu_prime_3_trick_details.aboutToShow.disconnect(self._create_trick_details_prime_3)
        self._setup_difficulties_menu(RandovaniaGame.PRIME3, self.menu_prime_3_trick_details)

    def _setup_difficulties_menu(self, game: RandovaniaGame, menu: QtWidgets.QMenu):
        game = default_database.game_description_for(game)
        tricks_in_use = used_tricks(game)

        menu.clear()
        for trick in sorted(game.resource_database.trick, key=lambda _trick: _trick.long_name):
            if trick not in tricks_in_use:
                continue

            trick_menu = QMenu(self)
            trick_menu.setTitle(trick.long_name)
            menu.addAction(trick_menu.menuAction())

            used_difficulties = difficulties_for_trick(game, trick)
            for i, trick_level in enumerate(iterate_enum(LayoutTrickLevel)):
                if trick_level in used_difficulties:
                    difficulty_action = QAction(self)
                    difficulty_action.setText(trick_level.long_name)
                    trick_menu.addAction(difficulty_action)
                    difficulty_action.triggered.connect(
                        functools.partial(self._open_trick_details_popup, game, trick, trick_level))

    # ==========

    @asyncSlot()
    async def _on_validate_seed_change(self):
        old_value = self._options.advanced_validate_seed_after
        new_value = self.menu_action_validate_seed_after.isChecked()

        if old_value and not new_value:
            box = QMessageBox(self)
            box.setWindowTitle("Disable validation?")
            box.setText(_DISABLE_VALIDATION_WARNING)
            box.setIcon(QMessageBox.Warning)
            box.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
            box.setDefaultButton(QMessageBox.No)
            user_response = await async_dialog.execute_dialog(box)
            if user_response != QMessageBox.Yes:
                self.menu_action_validate_seed_after.setChecked(True)
                return

        with self._options as options:
            options.advanced_validate_seed_after = new_value

    def _on_generate_time_limit_change(self):
        is_checked = self.menu_action_timeout_generation_after_a_time_limit.isChecked()
        with self._options as options:
            options.advanced_timeout_during_generation = is_checked

    def _on_menu_action_dark_mode(self):
        with self._options as options:
            options.dark_mode = self.menu_action_dark_mode.isChecked()

    def _open_auto_tracker(self):
        from randovania.gui.auto_tracker_window import AutoTrackerWindow
        self.auto_tracker_window = AutoTrackerWindow(common_qt_lib.get_game_connection(), self._options)
        self.auto_tracker_window.show()

    def _on_menu_action_previously_generated_games(self):
        path = self._options.data_dir.joinpath("game_history")
        try:
            if platform.system() == "Windows":
                os.startfile(path)
            elif platform.system() == "Darwin":
                subprocess.run(["open", path])
            else:
                subprocess.run(["xdg-open", path])
        except OSError:
            print("Exception thrown :)")
            box = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Information, "Game History",
                                        f"Previously generated games can be found at:\n{path}",
                                        QtWidgets.QMessageBox.Ok, self)
            box.setTextInteractionFlags(Qt.TextSelectableByMouse)
            box.show()

    def _on_menu_action_layout_editor(self):
        from randovania.gui.corruption_layout_editor import CorruptionLayoutEditor
        self.corruption_editor = CorruptionLayoutEditor()
        self.corruption_editor.show()

    def _update_hints_text(self):
        game_description = default_database.default_prime2_game_description()
        item_database = default_database.default_prime2_item_database()

        rows = []

        for item in item_database.major_items.values():
            rows.append((
                item.name,
                item.item_category.hint_details[1],
                item.item_category.general_details[1],
                item.broad_category.hint_details[1],
            ))

        from randovania.games.prime.echoes_items import DARK_TEMPLE_KEY_NAMES
        for dark_temple_key in DARK_TEMPLE_KEY_NAMES:
            rows.append((
                dark_temple_key.format("").strip(),
                ItemCategory.TEMPLE_KEY.hint_details[1],
                ItemCategory.TEMPLE_KEY.general_details[1],
                ItemCategory.KEY.hint_details[1],
            ))

        rows.append((
            "Sky Temple Key",
            ItemCategory.SKY_TEMPLE_KEY.hint_details[1],
            ItemCategory.SKY_TEMPLE_KEY.general_details[1],
            ItemCategory.KEY.hint_details[1],
        ))

        for item in item_database.ammo.values():
            rows.append((
                item.name,
                ItemCategory.EXPANSION.hint_details[1],
                ItemCategory.EXPANSION.general_details[1],
                item.broad_category.hint_details[1],
            ))

        self.hint_item_names_tree_widget.setRowCount(len(rows))
        for i, elements in enumerate(rows):
            for j, element in enumerate(elements):
                self.hint_item_names_tree_widget.setItem(i, j, QtWidgets.QTableWidgetItem(element))

        for i in range(4):
            self.hint_item_names_tree_widget.resizeColumnToContents(i)

        number_for_hint_type = {
            hint_type: i + 1
            for i, hint_type in enumerate(LoreType)
        }
        used_hint_types = set()

        self.hint_tree_widget.setSortingEnabled(False)

        # TODO: This ignores the Dark World names. But there's currently no logbook nodes in Dark World.
        for world in game_description.world_list.worlds:

            world_item = QtWidgets.QTreeWidgetItem(self.hint_tree_widget)
            world_item.setText(0, world.name)
            world_item.setExpanded(True)

            for area in world.areas:
                hint_types = {}

                for node in area.nodes:
                    if isinstance(node, LogbookNode):
                        if node.required_translator is not None:
                            hint_types[node.lore_type] = node.required_translator.short_name
                        else:
                            hint_types[node.lore_type] = "✓"

                if hint_types:
                    area_item = QtWidgets.QTreeWidgetItem(world_item)
                    area_item.setText(0, area.name)

                    for hint_type, text in hint_types.items():
                        area_item.setText(number_for_hint_type[hint_type], text)
                        used_hint_types.add(hint_type)

        self.hint_tree_widget.resizeColumnToContents(0)
        self.hint_tree_widget.setSortingEnabled(True)
        self.hint_tree_widget.sortByColumn(0, QtCore.Qt.AscendingOrder)

        for hint_type in used_hint_types:
            self.hint_tree_widget.headerItem().setText(number_for_hint_type[hint_type], hint_type.long_name)



import multiprocessing
from distutils.version import StrictVersion
from pathlib import Path

try:
    import nod
except ImportError:
    nod = None

from randovania.interface_common.game_workdir import validate_game_files_path
from randovania.interface_common.status_update_lib import ProgressUpdateCallable


def _disc_unpack_process(output_pipe,
                         iso: Path,
                         game_files_path: Path,
                         ):
    def progress_callback(path, progress):
        output_pipe.send((False, path, progress))

    def _helper():
        try:
            disc, is_wii = nod.open_disc_from_image(str(iso))
            data_partition = disc.get_data_partition()
            if not data_partition:
                raise RuntimeError(
                    "Could not find a data partition in '{}'.\nIs it a valid Metroid Prime 2 ISO?".format(
                        iso))

            context = nod.ExtractionContext()
            context.set_progress_callback(progress_callback)
            data_partition.extract_to_directory(str(game_files_path), context)
            return True, None, 100

        except RuntimeError as e:
            return True, str(e), 0

    output_pipe.send(_helper())


def _disc_pack_process(status_queue,
                       iso: Path,
                       game_files_path: Path,
                       ):
    def fprogress_callback(progress: float, name: str, bytes: int):
        status_queue.send((False, name, progress))

    def _helper():
        try:
            disc_builder = nod.DiscBuilderGCN(str(iso), fprogress_callback)
            disc_builder.build_from_directory(str(game_files_path))
            return True, None, 100

        except RuntimeError as e:
            return True, str(e), 0

    status_queue.send(_helper())


def _shared_process_code(target,
                         iso: Path,
                         game_files_path: Path,
                         on_finish_message: str,
                         progress_update: ProgressUpdateCallable):
    receiving_pipe, output_pipe = multiprocessing.Pipe(False)
    process = multiprocessing.Process(
        target=target,
        args=(output_pipe, iso, game_files_path)
    )
    process.start()
    try:
        finished = False
        message = ""
        percentage = 0
        while not finished:
            progress_update(message, percentage)
            finished, message, percentage = receiving_pipe.recv()

        if isinstance(message, str):
            raise RuntimeError(message)

        progress_update(on_finish_message, 1)
    finally:
        process.terminate()


def unpack_iso(iso: Path,
               game_files_path: Path,
               progress_update: ProgressUpdateCallable,
               ):
    try:
        game_files_path.mkdir(parents=True, exist_ok=True)
    except OSError as e:
        raise RuntimeError("Unable to create files dir {}:\n{}".format(
            game_files_path, e))

    _shared_process_code(
        target=_disc_unpack_process,
        iso=iso,
        game_files_path=game_files_path,
        on_finish_message="Finished extracting ISO",
        progress_update=progress_update
    )


def pack_iso(iso: Path,
             game_files_path: Path,
             progress_update: ProgressUpdateCallable
             ):
    validate_game_files_path(game_files_path.joinpath("files"))

    nod_version = StrictVersion(getattr(nod, "VERSION", "0.0.0"))
    if nod_version < StrictVersion("1.1.0"):
        raise RuntimeError("Installed nod version ({}) is older than required 1.1.0".format(nod_version))

    if nod.DiscBuilderGCN.calculate_total_size_required(str(game_files_path)) is None:
        raise RuntimeError("Image built with given directory would pass the maximum size.")

    _shared_process_code(
        target=_disc_pack_process,
        iso=iso,
        game_files_path=game_files_path,
        on_finish_message="Finished packing ISO",
        progress_update=progress_update
    )


def can_process_iso() -> bool:
    return nod is not None
